/**
  @file ComponentManager.c 
  @brief Routines for managing component services
*/
/*
   Copyright (C)2004 Matthew Bishop, Anthony Arcieri
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include "ComponentManager.h"
#include "xmalloc.h"

#include <assert.h>
#include <string.h>

struct _Component {
	ComponentRoutine init;
	ComponentRoutine reinit;
	ComponentRoutine shutdown;
};

/* Dependencies are kept in an array immediately after the struct in the same
   malloc()'d block. Access the array using the component_deps() macro.
 */
struct DepInfo {
	int visited;
	int n_deps;
};

#define component_deps(di) ((Component *)(((struct DepInfo *)di) + 1))
#define depinfo_size(n_deps) (sizeof(struct DepInfo) + sizeof(struct DepInfo *) * (n_deps))

static Component *components;
static struct DepInfo **depinfo;
static int n_components;

#ifndef NDEBUG
/* true after componentmgr_prepare_traversal() has been called. */
static int component_addition_complete;
#endif

Component component_create(ComponentRoutine init, ComponentRoutine reinit, ComponentRoutine shutdown)
{
	struct DepInfo *di;
	Component c;

	assert(!component_addition_complete);

	c = NEW(Component);
	c->init = init;
	c->reinit = reinit;
	c->shutdown = shutdown;

	di = (struct DepInfo *)xmalloc(sizeof(struct DepInfo));
	di->visited = 0;
	di->n_deps = 0;

	components = (Component *)xrealloc(components, sizeof(Component) * (n_components + 1));
	depinfo = (struct DepInfo **)xrealloc(depinfo, sizeof(struct DepInfo *) * (n_components + 1));
	components[n_components] = c;
	depinfo[n_components] = di;
	n_components++;

	return c;
}

static int component_index(Component c)
{
	int i;

	for(i = 0; i < n_components; i++) {
		if(components[i] == c)
			return i;
	}
	assert(i < n_components);

	return -1;
}

void component_register_dependency(Component module, Component dependency)
{
	int i = component_index(module);

	depinfo[i] = (struct DepInfo *)xrealloc(depinfo[i], depinfo_size(depinfo[i]->n_deps + 1));

	component_deps(depinfo[i])[depinfo[i]->n_deps] = dependency;
	depinfo[i]->n_deps++;
}

void componentmgr_init(void)
{
	components = NULL;
	depinfo = NULL;
	n_components = 0;
#ifndef NDEBUG
	component_addition_complete = 0;
#endif
}

void componentmgr_destroy(void)
{
	int i;

	for(i = 0; i < n_components; i++)
		xfree(components[i]);

	xfree(components);
}

static void inorder_visit(Component c, Component *heap, int *hi)
{
	struct DepInfo *di;
	int i, j;

	di = depinfo[component_index(c)];
	assert(di != NULL);

	/* visit each dependency in the order they were created */
	for(i = 0; i < n_components; i++) {
		for(j = 0; j < di->n_deps; j++) {
			if(components[i] == component_deps(di)[j]) {
				inorder_visit(components[i], heap, hi);
				break;
			}
		}
	}

	if(!di->visited) {
		heap[*hi] = c;
		(*hi)++;
		di->visited = 1;
	}
}

static void componentmgr_prepare_traversal(void)
{
	Component *roots, *deps, *heap;
	int i, j;

	assert(!component_addition_complete);

	roots = (Component *)xmalloc(sizeof(Component) * n_components);
	memcpy(roots, components, sizeof(Component) * n_components);

	/* remove all components that are deps of other components; the
	   remaining components are the roots */
	for(i = 0; i < n_components; i++) {
		deps = component_deps(depinfo[i]);
		for(j = 0; j < depinfo[i]->n_deps; j++)
			roots[component_index(deps[j])] = NULL;
	}

	/* traverse dependency tree, adding each component in-order */
	heap = (Component *)xmalloc(sizeof(Component) * n_components);

	for(i = j = 0; i < n_components; i++) {
		if(roots[i])
			inorder_visit(roots[i], heap, &j);
	}

	xfree(roots);
	xfree(components);
	components = heap;

	for(i = 0; i < n_components; i++)
		xfree(depinfo[i]);
	xfree(depinfo);

#ifndef NDEBUG
	component_addition_complete = 1;
#endif
}

void componentmgr_init_components(void)
{
	int i;

	if(!component_addition_complete)
		componentmgr_prepare_traversal();
	
	for(i = 0; i < n_components; i++) {
		if(components[i]->init)
			components[i]->init();
	}
}

void componentmgr_reinit_components(void)
{
	int i;

	if(!component_addition_complete)
		componentmgr_prepare_traversal();

	for(i = 0; i < n_components; i++) {
		if(components[i]->reinit)
			components[i]->reinit();
	}
}

void componentmgr_shutdown_components(void)
{
	int i;

	if(!component_addition_complete)
		componentmgr_prepare_traversal();

	for(i = n_components - 1; i >= 0; i--) {
		if(components[i]->shutdown)
			components[i]->shutdown();
	}
}
