#ifndef COMPONENTMANAGER_H
#define COMPONENTMANAGER_H

/**
   @file ComponentManager.h 
   @brief A startup component dependency manager
*/
/** 
 * @defgroup componentmanager Component Manager
 * @{
 */

/** Function signature for component init, reinit, and shutdown callbacks. */
typedef void (*ComponentRoutine)(void);

/** Component data type. */
typedef struct _Component *Component;

/** Initialize the component dependency manager. This must be called before
   any other component functions. */

void componentmgr_init(void);
/** Destroys the component dependency manager and all the components that
   have been created. This must be called at the end of the program and not
   be followed by other component function calls. */

void componentmgr_destroy(void);

/** Calls the init routine for all components in an order such that the
   components that are depended upon by a given component have already
   been initialized. */
void componentmgr_init_components(void);

/** Calls the reinit routine for all components in an order such that the
   components that are depended upon by a given component have already
   been reinitialized. */
void componentmgr_reinit_components(void);

/** Calls the shutdown routine for all components in an order such that
   a given component is shut down before the components it depends on. */
void componentmgr_shutdown_components(void);

/** Creates a new component with the given initialization, reinitialization,
   and shutdown callback routines. Such a component can be put in a
   dependency graph. so that each components callback is called in the
   correct order.
   @returns the new component
   @param init the initialization routine
   @param reinit the reinitialization routine
   @param shutdown the shutdown routine
   @see component_register_dependency */
Component component_create(ComponentRoutine init, ComponentRoutine reinit, ComponentRoutine shutdown);

/** Creates a dependency where dependency needs to be (re)initialized
   before module and module be shutdown before dependency. */
void component_register_dependency(Component module, Component dependency);

/** @} */
#endif /* COMPONENTMANAGER_H */
