/**
  @file Config.c
  @brief Configuration accessor methods and file parser
*/
/*
   Copyright (C)2003-04 Matthew Bishop, Anthony Arcieri
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "Config.h"
#include "HashTable.h"
#include "Lock.h"
#include "log.h"
#include "xmalloc.h"

#ifdef WIN32
#define strcasecmp(s1, s2)	_stricmp(s1, s2)
#endif

#define LINE_BUFFER_SIZE 512

static HashTable conf_table = 0;
static Lock conf_lock = 0;

/* special state abbreviations for the table */
#define SCAN_ERROR    (-4)
#define SKIP_CHAR     (-3)
#define DONE_SCANNING (-2)
#define ACCEPT_TOKEN  (-1)
#define START_STATE     0
#define AX ACCEPT_TOKEN
#define DN DONE_SCANNING
#define SK SKIP_CHAR
#define SE SCAN_ERROR

/* token types */
#define TOK_ERROR  (-1)
#define TOK_DONE    (0)
#define TOK_WORD    (1)
#define TOK_COMMENT (2)
#define TOK_STRING  (3)
#define TOK_SECTION (4)
#define TOK_EQUALS  (5)
#define TOK_COMMA   (6)
#define TOK_NEWLINE (7)

/* The Transition Table
 *
 * 'Special states' handle exceptional scan conditions, including accepting a
 * token. 'Regular states' are what the scanner is in most of the time; some
 * are final. 'Implicit accept states' always accept, and always return to
 * the start state, though they are not listed in the state transition table.
 *
 * character classes:
 *   0 - white space
 *   1 - letters
 *   2 - "
 *   3 - '
 *   4 - ,
 *   5 - =
 *   6 - [
 *   7 - ]
 *   8 - #
 *   9 - \n
 *  10 - EOF
 *
 * special states:
 *   -4 - scan error     (SE)
 *   -3 - skip character (SK)
 *   -2 - done parsing   (DN)
 *   -1 - accept token   (AX)
 * regular states:
 *    0 - start state
 *    1 - word
 *    2 - " str
 *    3 - ' str
 *    4 - section
 *    5 - comment
 * implicit accept states:
 *    6 - accept string
 *    7 - accept ','
 *    8 - accept '='
 *    9 - accept section
 *   10 - accept newline
 */
static const signed char transition[][11] = {
/* WS LT  "  '  ,  =  [   ]  # \n EOF */
  {SK, 1, 2, 3, 7, 8, 4,SE, 5,10,DN},      /* 0 start state */
  {AX, 1,SE,SE,AX,AX,SE,SE,AX,AX,AX},      /* 1 word */
  { 2, 2, 6, 2, 2, 2, 2, 2, 2,SE,SE},      /* 2 " string */
  { 3, 3, 3, 6, 3, 3, 3, 3, 3,SE,SE},      /* 3 ' string */
  {SK, 4,SE,SE,SE,SE,SE, 9,SE,SE,SE},      /* 4 section */
  {SK,SK,SK,SK,SK,SK,SK,SK,SK,AX,AX}       /* 5 comment */
};

/* state => token type table */
static const char state_type[] = {
	TOK_DONE, TOK_WORD, TOK_ERROR, TOK_ERROR, TOK_ERROR, TOK_COMMENT,
	TOK_STRING, TOK_COMMA, TOK_EQUALS, TOK_SECTION, TOK_NEWLINE
};

/* Parse state. Static vars work because the config parser acquires an
   exclusive lock. */
static char *config_filename = 0;
static char **config_keys;
static int config_line;
static int state;
static int toktype;
static char *tokbuf;
static char *section;

static int next_char(FILE *conffile)
{
	int c = fgetc(conffile);

	if(c == '\r')
		c = fgetc(conffile);

	if(c == '\n')
		config_line++;

	return c;
}

static void unget_char(int c, FILE *conffile)
{
	ungetc(c, conffile);

	if(c == '\n')
		config_line--;
}

/* Returns the class of the given character. */
static int class_of(int c)
{
	switch(c) {
	case EOF:
		return 10;
	case '"':
		return 2;
	case '\'':
		return 3;
	case ',':
		return 4;
	case '=':
		return 5;
	case '[':
		return 6;
	case ']':
		return 7;
	case '#':
		return 8;
	case '\n':
		return 9;
	default:
		if(isspace(c))
			return 0;
		break;
	}
	return 1;
}

static int get_token(FILE *conffile)
{
	int i = 0, c, newstate;

	state = START_STATE;

	for(;;) {
		c = next_char(conffile);
		tokbuf[i] = c;
		newstate = transition[state][class_of(c)];

		switch(newstate) {
		case SCAN_ERROR:
			log_printf("!!! %s:%i: scan error; skipping entire line", config_filename, config_line);
			goto skipline;
		case SKIP_CHAR:
			break;
		case ACCEPT_TOKEN:
			unget_char(c, conffile);
			tokbuf[i] = '\0';
		case DONE_SCANNING:
			return state_type[state];
		default:
			if(newstate >= 6) { /* implicit accept state */
				tokbuf[i] = '\0';
				return state_type[newstate];
			}
			i++;
			state = newstate;
			break;
		}

		if(i == LINE_BUFFER_SIZE) {
			log_printf("!!! %s:%i: token too long; skipping entire line", config_filename, config_line);
			goto skipline;
		}
	}
skipline:
	do
		c = fgetc(conffile);
	while(c != EOF && c != '\n');

	return TOK_NEWLINE;
}

static int next_token(FILE *conffile)
{
	int type;

	do
		type = get_token(conffile);
	while(type != TOK_DONE && type == TOK_COMMENT);

	toktype = type;
	return type;
}

static int skip_tokens_until(FILE *conffile, int untiltype)
{
	int type;

	do
		type = next_token(conffile);
	while(type != TOK_DONE && type != untiltype);

	toktype = type;
	return type;
}

static int skip_tokens_while(FILE *conffile, int whiletype)
{
	int type;

	do
		type = next_token(conffile);
	while(type != TOK_DONE && type == whiletype);

	toktype = type;
	return type;
}

static int is_key(const char *key, const char *allowed_keys)
{
	size_t i, keylen;

	keylen = strlen(key);
	assert(keylen != 0);

	for(i = 0; *allowed_keys != '\0'; i++) {
		size_t len = strcspn(allowed_keys, ",");
		if(keylen == len && !strncmp(key, allowed_keys, len))
			return 1;
		allowed_keys += len + 1;
	}

	return 0;
}

static int is_name(const char *section_, const char *key)
{
	size_t i, section_len;

	section_len = strlen(section_);

	for(i = 0; config_keys[i] != 0; i++) {
		size_t len = strcspn(config_keys[i], ":");
		if(section_len == len && !strncmp(section_, config_keys[i], len)) {
			if(key != 0)
				return is_key(key, config_keys[i] + len + 1);
			else
				return 1;
		}
	}

	return 0;
}

static int is_section_name(const char *section_)
{
	return is_name(section_, 0);
}

static int is_key_name(const char *key)
{
	return is_name(section, key);
}

/* Free a value list within the config */
static void value_free(void *ptr)
{
	int i = 0;
	void **v = (void **)ptr;

	while(v[i])
		xfree(v[i++]);

	xfree(v);
}

/* Free each section of a hash table */
static void section_free(void *ptr)
{
	hash_table_destroy((HashTable)ptr, value_free);
}

static void add_value(char *varname, char **values)
{
	HashTable sect_table;

	if(!(sect_table = (HashTable)hash_table_lookup(conf_table, section))) {
		sect_table = hash_table_create();
		hash_table_insert(conf_table, section, sect_table);
	}

	hash_table_insert(sect_table, varname, values);
}

static int parse_section(FILE *conffile)
{
	if(toktype == TOK_DONE)
		return -1;

	assert(toktype == TOK_SECTION);

	section = xstrdup(tokbuf + 1);

	if(!is_section_name(section)) {
		log_printf("!!! %s:%i: invalid section [%s]; skipping to next section", config_filename, config_line, section);
		skip_tokens_until(conffile, TOK_SECTION);
		return parse_section(conffile);
	}

	if(next_token(conffile) != TOK_NEWLINE) {
		log_printf("!!! %s:%i: expected newline after [%s]; ignoring to end of line", config_filename, config_line, section);
		skip_tokens_until(conffile, TOK_NEWLINE);
	}

	skip_tokens_while(conffile, TOK_NEWLINE);

	return 0;
}

static int parse_keyvalue(FILE *conffile)
{
	char *key, *value, **values = 0;
	unsigned nvalues = 0;

	if(toktype != TOK_WORD)
		return -1;

	/* get key */
	key = xstrdup(tokbuf);

	if(!is_key_name(key)) {
		log_printf("!!! %s:%i: invalid key '%s'; ignoring line", config_filename, config_line, key);
		skip_tokens_until(conffile, TOK_NEWLINE);
		goto done;
	}

	if(next_token(conffile) != TOK_EQUALS) {
		log_printf("!!! %s:%i: expected '=' after key; ignoring line", config_filename, config_line);
		skip_tokens_until(conffile, TOK_NEWLINE);
		goto done;
	}

	/* parse values */
	for(;;) {
		switch(next_token(conffile)) {
		case TOK_WORD:
		case TOK_STRING:
parse_value:
			nvalues++;
			values = (char **)xrealloc(values, (nvalues + 1) * sizeof(char *));

			value = tokbuf;
			if(toktype == TOK_STRING)
				value++;

			values[nvalues - 1] = xstrdup(value);
			break;
		case TOK_NEWLINE:
			if(nvalues > 0)
				goto addvalue;
		default:
			log_printf("!!! %s:%i: expected key = value; ignoring line", config_filename, config_line);
			skip_tokens_until(conffile, TOK_NEWLINE);
			goto done;
		}

		next_token(conffile);
		if(toktype == TOK_NEWLINE || toktype == TOK_DONE) {
addvalue:
			values[nvalues] = 0;
			add_value(key, values);
			goto done;
		} else if(toktype == TOK_WORD || toktype == TOK_STRING) {
			log_printf("!!! %s:%i: expected ',' between values", config_filename, config_line);
			goto parse_value;
		} else if(toktype != TOK_COMMA) {
			log_printf("!!! %s:%i: syntax error; ignoring line", config_filename, config_line);
			skip_tokens_until(conffile, TOK_NEWLINE);
			goto done;
		}
	}

done:
	xfree(key);

	skip_tokens_while(conffile, TOK_NEWLINE);

	return 0; /* keep looking for key = value lines */
}

static void parse_config(FILE *conffile)
{
	while(next_token(conffile) == TOK_NEWLINE) ;

	if(toktype != TOK_SECTION) {
		log_printf("!!! %s:%i: expected [section] in config file; ignoring", config_filename, config_line);

		do { /* find first [section] */
			do
				if(next_token(conffile) == TOK_DONE)
					return;
			while(toktype != TOK_NEWLINE);

			if(next_token(conffile) == TOK_DONE)
				return;
		} while(toktype != TOK_SECTION);
	}

	for(;;) {
		if(section != 0)
			xfree(section);

		if(parse_section(conffile) == -1)
			return;

		while(parse_keyvalue(conffile) == 0) ;
	}
}

int config_init(const char *filename, const char **keys)
{
	FILE *conffile;

	if(!conf_lock)
		conf_lock = lock_create(LOCK_SHARED);

	lock_writer_acquire(conf_lock);

	if(!(conffile = fopen(filename, "r"))) {
		lock_writer_release(conf_lock);
		return -1;
	}

	if(conf_table != 0)
		hash_table_destroy(conf_table, section_free);

	conf_table = hash_table_create();

	tokbuf = (char *)xmalloc(LINE_BUFFER_SIZE);

	if(config_filename)
		xfree(config_filename);

	config_filename = xstrdup(filename);
	config_keys = (char **)keys;
	config_line = 1;
	state = START_STATE;
	section = 0;

	parse_config(conffile);

	xfree(tokbuf);
	fclose(conffile);

	lock_writer_release(conf_lock);

	return 0;
}

void config_destroy()
{
	if(config_filename)
		xfree(config_filename);

	if(conf_table)
		hash_table_destroy(conf_table, section_free);

	if(conf_lock)
		lock_destroy(conf_lock);
}

static char **config_vlist(const char *section, const char *var)
{
	HashTable sect_table;

	if(!conf_table)
		return 0;

	if(!(sect_table = (HashTable)hash_table_lookup(conf_table, section)))
		return 0;

	return (char **)hash_table_lookup(sect_table, var);
}

char *config_vlist_value(const char *section, const char *var, int i)
{
	char **vlist, *ret = 0;

	lock_reader_acquire(conf_lock);
	if((vlist = config_vlist(section, var)) && vlist[i])
		ret = xstrdup(vlist[i]);

	lock_reader_release(conf_lock);
	return ret;
}

char *config_value(const char *section, const char *var)
{
	char **vlist, *ret = 0;

	lock_reader_acquire(conf_lock);
	if((vlist = config_vlist(section, var)) && vlist[0])
		ret = xstrdup(vlist[0]);

	lock_reader_release(conf_lock);
	return ret;
}

int config_int_value(const char *section, const char *var)
{
	int ret = -1;
	char **vlist;

	lock_reader_acquire(conf_lock);
	if((vlist = config_vlist(section, var)) && vlist[0])
		ret = atoi(vlist[0]);

	lock_reader_release(conf_lock);

	return ret;
}

int config_truth_value(const char *section, const char *var)
{
	int ret = -1;
	char **vlist, *value;

	lock_reader_acquire(conf_lock);
	if(!(vlist = config_vlist(section, var)) || !vlist[0])
		goto done;

	ret = 1;
	value = vlist[0];

	if(!strcasecmp(value, "true"))
		goto done;

	if(!strcasecmp(value, "yes"))
		goto done;

	if(strlen(value) == 1 && (*value == 'y' || *value == 'Y'))
		goto done;

	ret = 0;
done:
	lock_reader_release(conf_lock);
	return ret;
}
