/**
  @file DirEntry.c
  @brief Abstract filesystem entity information querying interface
 */
/*
   Copyright (C)2004 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <string.h>
#ifndef WIN32
#include <sys/stat.h>
#include <time.h>
#endif

#include "DirEntry.h"
#include "debug.h"
#include "log.h"
#include "global.h"
#include "util.h"
#include "xmalloc.h"

#ifdef WIN32
void trim_trailing_slash(char *destpath, const char *srcpath, unsigned destlen)
{
	size_t srclen = strlen(srcpath);

	if(srclen < 1 || srcpath[srclen - 1] != '\\') {
		safecpy(destpath, srcpath, destlen);
		return;
	}

	safecpy(destpath, srcpath, srclen > destlen ? destlen : (unsigned)srclen);
}
#endif /* WIN32 */

int direntry_get(const char *path, struct DirEntry *entry)
{
	char *t;
	
#ifdef WIN32
	char realpath[PDTP_PATH_MAX];
	HANDLE find;
	WIN32_FIND_DATA data;
    LARGE_INTEGER fsize, ctime, yeartime;
    SYSTEMTIME systime;
    FILETIME ftime;
#else
    int r, fd;
    struct stat st;
	struct tm tm;
#endif

#ifdef WIN32
	/* Trim the trailing slash on directory names */
	trim_trailing_slash(realpath, path, PDTP_PATH_MAX);

	/* Open to ensure readability */
	if((find = FindFirstFile(realpath, &data)) == INVALID_HANDLE_VALUE) {
		log_error(realpath);
		return -1;
	}

	CloseHandle(find);

	if(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
		entry->type = ENTRY_DIR;
		entry->length = 0;
	} else {
                entry->type = ENTRY_FILE;

                fsize.HighPart = data.nFileSizeHigh;
                fsize.LowPart = data.nFileSizeLow;
                entry->length = fsize.QuadPart;
	}

	if(!(t = strrchr(path, '\\')))
		safecpy(entry->name, path, PDTP_PATH_MAX);
	else 
		safecpy(entry->name, ++t, PDTP_PATH_MAX);

    FileTimeToSystemTime(&data.ftLastWriteTime, &systime);

    entry->ts_year = systime.wYear;

    /* Set values relative to January 1st */
    systime.wMonth = 1;
    systime.wDay = 1;
    systime.wHour = 0;
    systime.wMinute = 0;
    systime.wSecond = 0;
    systime.wMilliseconds = 0;

    SystemTimeToFileTime(&systime, &ftime);

    yeartime.HighPart = ftime.dwHighDateTime;
    yeartime.LowPart = ftime.dwLowDateTime;

    ctime.HighPart = data.ftLastWriteTime.dwHighDateTime;
    ctime.LowPart = data.ftLastWriteTime.dwLowDateTime;

    entry->ts_ysec = (int)((ctime.QuadPart - yeartime.QuadPart) / 10000000);
#else
	/* Open to ensure readability */		
    if((fd = open(path, O_RDONLY)) < 0)
		return -1;

        r = fstat(fd, &st);
        close(fd);

        if(r < 0)
                return -1;

        if(S_ISDIR(st.st_mode)) {
                entry->type = ENTRY_DIR;
                entry->length = 0;
        } else {
                entry->type = ENTRY_FILE;
                entry->length = st.st_size;
        }

        if(!(t = strrchr(path, '/')))
		safecpy(entry->name, path, PDTP_PATH_MAX);
	else 
		safecpy(entry->name, ++t, PDTP_PATH_MAX);

        localtime_r(&st.st_mtime, &tm);
        entry->ts_year = tm.tm_year + 1900;
        entry->ts_ysec = tm.tm_yday * 86400 + tm.tm_hour * 3600 + tm.tm_min * 60
 + tm.tm_sec;
#endif /* WIN32 */

        return 0;
}
