/**
  @file Dispatcher.c 
  @brief Dispatches incoming socket events to worker threads

NOTICE: Some refactoring is probably in order on all of this code.  

Originally common coupling was used in order to let one thread manipulate 
the dispatcher data structures while the dispatcher thread was running.
Unfortunately this approach was plagued by race conditions.  To amend this,
a pipe is used to notify the dispatcher thread of additions to or removals
from the dispatcher data structures.  However, the data structures are
still implemented as global static variables.  This does have the advantage
that since these variables are initialized by dispatcher_init function
problems initializing the dispatcher can be caught before a new thread
is ever spawned, however it is considered bad style by many to use globally
scoped variables.

A QUICK EXPLANATION OF THESE FUNCTIONS:

The purpose of these functions is to use the best available event monitoring
mechanism available on the underlying platform to receive socket events, then
dispatch those socket events to a worker thread which then invokes a specified
handler function.  Every implementation will contain the main event loop for
the server, which will block indefinitely until an event is received.  Each
implementation is capable of receiving multiple event notifications within
a system call, including the three stateful mechanisms implemented here,
kqueues, /dev/poll, and epoll().  Due to the transactional nature of the
server, any of these implemenentations, with the exception of select(),
should be able to scale to thousands of active connections.
 */
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
 */
#include <machdep.h>

/* Linux 2.5/2.6 kernels with a recent glibc may want to try uncommenting the 
   following to use epoll() in place of poll() */  

/* #define HAVE_EPOLL */

#ifndef WIN32
#include <pthread.h>

#include "ExecQueue.h"
#include "WorkerManager.h"
#endif /* WIN32 */

#include "Dispatcher.h"
#include "msgval.h"
#include "debug.h"
#include "log.h"

/* Win32 I/O completion ports dispatcher */
#ifdef WIN32
#define HAVE_DISPATCHER

#define THREAD_COUNT	64

HANDLE cport = 0;
HANDLE threads[THREAD_COUNT];

static DWORD dispatcher_thread(LPVOID param)
{
	return 0;
}

void dispatcher_init(void)
{
	int i;

	if(!(cport = CreateIoCompletionPort(INVALID_HANDLE_VALUE, 0, 0, 0)))
		fatal("Couldn't create completion port");

	for(i = 0; i < THREAD_COUNT; i++) {
		if(!(threads[i] = CreateThread(0, 0, (LPTHREAD_START_ROUTINE)dispatcher_thread, 0, 0, 0)))
			fatal("Could spawn entire thread pool");
	}
}

void dispatcher_destroy(void)
{
}

void dispatcher_add(int fd, void (*event_handler)(msgval *arg))
{
}

void dispatcher_remove(int fd)
{
}
#endif /* WIN32 */

/* kqueue()-based implementation.  While MacOS X Panther supports kqueues, this
   implementation does not seem to function correctly on it, and consequently
   poll is used on MacOS X for the time being */
#if defined(HAVE_KQUEUE) && !defined(MACOSX) && !defined(HAVE_DISPATCHER)
#define HAVE_DISPATCHER

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "log.h"

/* Maximum number of events to read in a single system call */
#define KQUEUE_EVENTS	16

static int kq = -1;
static pthread_t dispatcher_tid;

static void *dispatcher_main(void *arg)
{
	int i, n;
	struct kevent kev[KQUEUE_EVENTS];
	msgval val;

	for(;;) {
		if((n = kevent(kq, 0, 0, kev, KQUEUE_EVENTS, 0)) <= 0) {
			perror("kevent");
			fatal("!!! Dispatcher failure");
		}

		for(i = 0; i < n; i++) {
			if(kev[i].flags & EV_ERROR)
				continue;

			val.num = kev[i].ident;
			workermgr_exec((void (*)(msgval *))kev[i].udata, &val);
		}
	}

	pthread_exit(0);
}

void dispatcher_init(void)
{
	int ret;

	if(kq != -1)
		return;

	if((kq = kqueue()) < 0) {
		perror("kqueue");
		fatal("!!! Dispatcher failure");
	}

	if((ret = pthread_create(&dispatcher_tid, 0, dispatcher_main, 0)))
		fatal("!!! Couldn't create dispatcher thread: %s\n", strerror(ret));
}

void dispatcher_destroy(void)
{
	/* XXX STUB! */
}

void dispatcher_add(int fd, void (*event_handler)(msgval *))
{
	struct kevent kev;

	EV_SET(&kev, fd, EVFILT_READ, EV_ADD | EV_ENABLE | EV_ONESHOT, 0, 0, (void *)event_handler);

	if(kevent(kq, &kev, 1, 0, 0, 0) < 0) {
		log_error("kevent");
		fatal("!!! Dispatcher failure");
	}
}

void dispatcher_remove(int fd)
{
	struct kevent kev;

	EV_SET(&kev, fd, EVFILT_READ, EV_DELETE, 0, 0, 0);

	if(kevent(kq, &kev, 1, 0, 0, 0) < 0) {
		log_error("kevent");
		fatal("!!! Dispatcher failure");
	}
}
#endif /* HAVE_KQUEUE */

/* sys_epoll based implementation - should work with Linux 2.6 systems */
#if defined(HAVE_EPOLL) && !defined(HAVE_DISPATCHER)
#define HAVE_DISPATCHER

#include <sys/epoll.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "log.h"

/* Value passed to epoll_create() */
#define EPOLL_SIZE_HINT		64

/* Number of events to receive in a single system call */
#define EPOLL_EVENTS		16

static int ep_fd = -1;
pthread_t dispatcher_tid;

void *dispatcher_main(void *arg)
{
	int i, n;
	struct epoll_event events[EPOLL_EVENTS];
	msgval val;

	for(;;) {
		if((n = epoll_wait(ep_fd, events, EPOLL_EVENTS, -1)) < 0) {
			perror("epoll_wait");
			fatal("!!! Dispatcher failure");
		}

		for(i = 0; i < n; i++) {
			if(epoll_ctl(ep_fd, EPOLL_CTL_DEL, events[i].data.fd, 0) < 0) {
				perror("epoll_ctl");
				fatal("!!! Dispatcher failure");
			}

			val.num = events[i].data.fd;
			workermgr_exec((void (*)(msgval *))events[i].data.ptr, &val);
		}
	}
}

void dispatcher_init(void)
{
	int ret;

	if((ep_fd = epoll_create(EPOLL_SIZE_HINT)) < 0) {
		perror("epoll_create");
		fatal("!!! Dispatcher failure");
	}

	if((ret = pthread_create(&dispatcher_tid, 0, dispatcher_main, 0)))
		fatal("!!! Couldn't create dispatcher thread: %s\n", strerror(ret));

}

void dispatcher_destroy(void)
{
	close(ep_fd);
}

void dispatcher_add(int sock, void (*event_handler)(msgval *))
{
	struct epoll_event ev;

	ev.events = EPOLLIN;
	ev.data.fd = sock;
	ev.data.ptr = event_handler;

	epoll_ctl(ep_fd, EPOLL_CTL_ADD, sock, &ev);
}

void dispatcher_remove(int sock)
{
	epoll_ctl(ep_fd, EPOLL_CTL_DEL, sock, 0);
}
#endif /* EVSOURCE_EPOLL */

/* /dev/poll based dispatcher implementation, for Solaris, Irix, HP-UX, etc. */
#ifdef HAVE_DEVPOLL

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/devpoll.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "Collection.h"
#include "log.h"

/* Maximum number of events to read in a single system call */
#define DEVPOLL_EVENTS	16

enum { 
	EVSOURCE_ADD,
	EVSOURCE_DEL,
	DISPATCHER_QUIT
};

struct dispatcher_op {
	int opcode;
	int fd;
	void (*event_handler)(msgval *arg);
};

static int dp_fd = -1;
static int dispatcher_infd, dispatcher_outfd;
static pthread_t dispatcher_tid;
static Collection function_table;

static void *dispatcher_main(void *arg)
{
	int i, n;
	struct dvpoll dp;
	struct pollfd pfd, inevents[DEVPOLL_EVENTS];
	struct dispatcher_op op;
	void (*event_handler)(msgval *);
	msgval val;

	dp.dp_fds = inevents;
	dp.dp_nfds = 16;
	dp.dp_timeout = -1;

	for(;;) {
		if((n = ioctl(dp_fd, DP_POLL, &dp)) < 1) {
			perror("/dev/poll");
			fatal("!!! Dispatcher failure");
		}

		for(i = 0; i < n; i++) {
			if(inevents[i].fd == dispatcher_infd) {
				if(read(dispatcher_infd, &op, sizeof(struct dispatcher_op)) < 0) {
					perror("read");
					fatal("!!! Dispatcher failure");
				}

				switch(op.opcode) {
					case EVSOURCE_ADD:
						collection_insert(function_table, op.fd, (void *)op.event_handler);

						pfd.fd = op.fd;
						pfd.events = POLLIN;
						pfd.revents = 0;

						write(dp_fd, &pfd, sizeof(struct pollfd));
						continue;
					case EVSOURCE_DEL:
						collection_remove(function_table, op.fd);

						pfd.fd = op.fd;
						pfd.events = POLLREMOVE;
						pfd.revents = 0;

						write(dp_fd, &pfd, sizeof(struct pollfd));
						continue;
					case DISPATCHER_QUIT:
						pthread_exit(0);
				}
			}

			if(!(event_handler = (void (*)(msgval *))collection_remove(function_table, inevents[i].fd)))
				fatal("!!! INTERNAL INCONSISTENCY: Dispatcher function table entry missing");

			val.num = inevents[i].fd;
			workermgr_exec(event_handler, &val);

			pfd.fd = inevents[i].fd;
			pfd.events = POLLREMOVE;
			pfd.revents = 0;

			write(dp_fd, &pfd, sizeof(struct pollfd));
		}
	}
}

void dispatcher_init(void)
{
	int ret, pipefd[2];
	struct pollfd pfd;

	if((dp_fd = open("/dev/poll", O_WRONLY)) < 0) {
		perror("/dev/poll");
		exit(EXIT_FAILURE);
	}

	if(pipe(pipefd) < 0) {
		perror("pipe");
		fatal("!!! Couldn't initialize dispatcher");
	}

	dispatcher_infd = pipefd[0];
	dispatcher_outfd = pipefd[1];

	pfd.fd = dispatcher_infd;
	pfd.events = POLLIN;
	pfd.revents = 0;

	write(dp_fd, &pfd, sizeof(struct pollfd));

	function_table = collection_create();

	if((ret = pthread_create(&dispatcher_tid, 0, dispatcher_main, 0)))
		fatal("!!! Couldn't create dispatcher thread: %s\n", strerror(ret));
}

void dispatcher_destroy(void)
{
	struct dispatcher_op op;

	/* Signal the dispatcher to quit */
	op.opcode = DISPATCHER_QUIT;
	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));

	/* Wait for the dispatcher to terminate */
	pthread_join(dispatcher_tid, 0);

	/* Free internal data structures */
	close(dp_fd);
	collection_destroy(function_table, 0);
}

void dispatcher_add(int fd, void (*event_handler)(msgval *arg))
{
	struct dispatcher_op op;

	op.opcode = EVSOURCE_ADD;
	op.fd = fd;
	op.event_handler = event_handler;

	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));
}

void dispatcher_remove(int fd)
{
	struct dispatcher_op op;

	op.opcode = EVSOURCE_DEL;
	op.fd = fd;

	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));
}

#endif /* HAVE_DEVPOLL */

/* poll()-based implementation */
#if defined(HAVE_POLL) && !defined(HAVE_DISPATCHER)
#define HAVE_DISPATCHER

#include <sys/poll.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "log.h"
#include "xmalloc.h"

/* Initial size of the file descriptor table */
#define INITIAL_TABLE_SIZE	64

typedef void (*EventHandler)(msgval *);

enum {
	EVSOURCE_ADD,
	EVSOURCE_DEL,
	DISPATCHER_QUIT
};

struct dispatcher_op {
	int opcode;
	int fd;
	EventHandler event_handler;
};

static int dispatcher_infd, dispatcher_outfd;
static pthread_t dispatcher_tid;
static struct pollfd *fd_table = 0;
static EventHandler *func_table = 0;
static int fd_table_count, fd_table_size;

static void *dispatcher_main(void *arg)
{
	int i, n;
	struct dispatcher_op op;
	msgval val;

	for(;;) {
		if((n = poll(fd_table, fd_table_count, -1)) < 0) {
			if(errno == EINTR)
				continue;

			perror("poll");
			fatal("!!! Dispatcher error");
		}

		if(fd_table[0].revents == POLLIN) {
			if(read(dispatcher_infd, &op, sizeof(struct dispatcher_op)) < 0) {
				perror("read");
				fatal("!!! Dispatcher failure");
			}

			switch(op.opcode) {
				case EVSOURCE_ADD:
					if(fd_table_count == fd_table_size) {
						fd_table_size *= 2;
						fd_table = (struct pollfd *)xrealloc(fd_table, fd_table_size * sizeof(struct pollfd));
						func_table = (EventHandler *)xrealloc(func_table, fd_table_size * sizeof(EventHandler));
					}

					fd_table[fd_table_count].fd = op.fd;
					fd_table[fd_table_count].events = POLLIN;
					fd_table[fd_table_count].revents = 0;

					func_table[fd_table_count] = op.event_handler;


					fd_table_count++;

					break;
				case EVSOURCE_DEL:
					for(i = 1; i < fd_table_count && fd_table[i].fd != op.fd; i++);

					if(i == fd_table_count)
						break;

					if(i < fd_table_count - 1) {
						memcpy(&fd_table[i], &fd_table[fd_table_count - 1], sizeof(struct pollfd));
						func_table[i] = func_table[fd_table_count - 1];
					}

					fd_table_count--;

					if(fd_table_size > INITIAL_TABLE_SIZE && fd_table_count < fd_table_size / 4) {
						fd_table_size /= 2;
						fd_table = (struct pollfd *)xrealloc(fd_table, fd_table_size * sizeof(struct pollfd));
						func_table = (EventHandler *)xrealloc(func_table, fd_table_size * sizeof(EventHandler));
					}

					break;
				case DISPATCHER_QUIT:
					pthread_exit(0);
			}

			n--;
		}

		i = 1;
		while(n && i < fd_table_count) {
			if(fd_table[i].revents != POLLIN) {
				i++;
				continue;
			}

			val.num = fd_table[i].fd;
			workermgr_exec(func_table[i], &val);

			if(i < fd_table_count - 1) {
				memcpy(&fd_table[i], &fd_table[fd_table_count - 1], sizeof(struct pollfd));
				func_table[i] = func_table[fd_table_count - 1];
			}

			fd_table_count--;

			if(fd_table_size > INITIAL_TABLE_SIZE && fd_table_count < fd_table_size / 4) {
				fd_table_size /= 2;
				fd_table = (struct pollfd *)xrealloc(fd_table, fd_table_size * sizeof(struct pollfd));
				func_table = (EventHandler *)xrealloc(func_table, fd_table_size * sizeof(EventHandler));
			}

			n--;
		}
	}
}

void dispatcher_init(void)
{
	int ret, pipefd[2];

	if(pipe(pipefd) < 0) {
		perror("pipe");
		fatal("!!! Couldn't initialize dispatcher");
	}

	dispatcher_infd = pipefd[0];
	dispatcher_outfd = pipefd[1];

	fd_table_size = INITIAL_TABLE_SIZE;
	fd_table = (struct pollfd *)xmalloc(fd_table_size * sizeof(struct pollfd));
	func_table = (EventHandler *)xmalloc(fd_table_size * sizeof(EventHandler));

	fd_table_count = 1;
	fd_table[0].fd = dispatcher_infd;
	fd_table[0].events = POLLIN;
	fd_table[0].revents = 0;

	if((ret = pthread_create(&dispatcher_tid, 0, dispatcher_main, 0)))
		fatal("!!! Couldn't create dispatcher thread: %s\n", strerror(ret));
}

void dispatcher_destroy(void)
{
	struct dispatcher_op op;

	/* Signal the dispatcher to quit */
	op.opcode = DISPATCHER_QUIT;
	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));

	/* Wait for the dispatcher to terminate */
	pthread_join(dispatcher_tid, 0);

	/* Free internal data structures */
	xfree(fd_table);
	xfree(func_table);
}

void dispatcher_add(int fd, void (*event_handler)(msgval *arg))
{
	struct dispatcher_op op;

	op.opcode = EVSOURCE_ADD;
	op.fd = fd;
	op.event_handler = event_handler;

	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));
}

void dispatcher_remove(int fd)
{
	struct dispatcher_op op;

	op.opcode = EVSOURCE_DEL;
	op.fd = fd;

	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));
}

#endif /* HAVE_POLL */

/* select()-based implementation */
#if defined(HAVE_SELECT) && !defined(HAVE_DISPATCHER)
#define HAVE_DISPATCHER

#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "xmalloc.h"
#include "debug.h"
#include "log.h"

/* To reduce heap usage and fragmentation, we store and recycle dispatcher
   frames.  This defines how many frames to create initially.  If we run out 
   of spare frames we will create more */
#define INITIAL_SPARE_FRAMES    1

/* This designates the maximum number of spare frames to keep on hand */
#define MAX_SPARE_FRAMES        16

enum {
	EVSOURCE_ADD,
	EVSOURCE_DEL,
	DISPATCHER_QUIT
};

struct dispatcher_op {
	int opcode;
	int fd;
	void (*event_handler)(msgval *arg);
};

typedef struct _MplxFrame {
	int fd;
	void (*event_handler)(msgval *arg);

	struct _MplxFrame *link;
} *MplxFrame;

static int spare_frame_count = 0;
static MplxFrame head = 0, spare_frames = 0;
static int dispatcher_infd, dispatcher_outfd;
static pthread_t dispatcher_tid;

static void *dispatcher_main(void *arg)
{
	struct dispatcher_op op;
	int active_fds, max_fd = -1;
	fd_set fdset;
	MplxFrame cur, last, frame;
	msgval val;

	for(;;) {
		FD_ZERO(&fdset);
		FD_SET(dispatcher_infd, &fdset);
		max_fd = dispatcher_infd;

		for(cur = head; cur; cur = cur->link) {
			FD_SET(cur->fd, &fdset);

			if(cur->fd > max_fd)
				max_fd = cur->fd;
		}

		if((active_fds = select(max_fd + 1, &fdset, 0, 0, 0)) < 0) {
			perror("select");
			fatal("!!! Dispatcher failure");
		}

		if(FD_ISSET(dispatcher_infd, &fdset)) {
			if(read(dispatcher_infd, &op, sizeof(struct dispatcher_op)) < 0) {
				perror("read");
				fatal("!!! Dispatcher failure");
			}

			switch(op.opcode) {
				case EVSOURCE_ADD:
					if(!(frame = spare_frames))
						frame = NEW(MplxFrame);
					else {
						spare_frames = frame->link;
						spare_frame_count--;
					}

					frame->link = head;
					head = frame;

					frame->fd = op.fd;
					frame->event_handler = op.event_handler;
					break;
				case EVSOURCE_DEL:
					if(!head)
						break;

					for(cur = head; cur->link != 0 && cur->link->fd != op.fd; cur = cur->link);

					if(!cur->link) 
						break;

					frame = cur->link;
					cur->link = frame->link;

					if(spare_frame_count > MAX_SPARE_FRAMES)
						xfree(frame); 
					else {
						frame->link = spare_frames;
						spare_frames = frame;
						spare_frame_count++;
					}

					break;
				case DISPATCHER_QUIT:
					pthread_exit(0);
			}
		}

		cur = head;
		last = 0;

		while(active_fds > 0 && cur) {
			if(FD_ISSET(cur->fd, &fdset)) {
				active_fds--;

				if(!last)
					head = cur->link;
				else
					last->link = cur->link;

				val.num = cur->fd;
				workermgr_exec(cur->event_handler, &val);

				if(spare_frame_count > MAX_SPARE_FRAMES)
					xfree(cur);
				else {
					cur->link = spare_frames;
					spare_frames = cur;
					spare_frame_count++;
				}
			}

			last = cur;
			cur = cur->link;
		}
	}
}

void dispatcher_init(void)
{
	int ret, pipefd[2];
	MplxFrame frame;

	if(pipe(pipefd) < 0) {
		perror("pipe");
		fatal("!!! Couldn't initialize dispatcher");
	}

	dispatcher_infd = pipefd[0];
	dispatcher_outfd = pipefd[1];

	for(spare_frame_count = 0; spare_frame_count < INITIAL_SPARE_FRAMES;
			spare_frame_count++)
	{
		frame = NEW(MplxFrame);
		frame->link = spare_frames;
		spare_frames = frame;
	}

	if((ret = pthread_create(&dispatcher_tid, 0, dispatcher_main, 0)))
		fatal("!!! Couldn't create dispatcher thread: %s\n", strerror(ret));
}

void dispatcher_destroy(void)
{
	struct dispatcher_op op;
	MplxFrame frame;

	/* Signal the dispatcher to quit */
	op.opcode = DISPATCHER_QUIT;
	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));

	/* Wait for the dispatcher to terminate */
	pthread_join(dispatcher_tid, 0);

	/* Free internal data structures */
	while(head) {
		frame = head;
		head = frame->link;

		xfree(frame);
	}

	while(spare_frames) {
		frame = spare_frames;
		spare_frames = frame->link;

		xfree(frame);
	}
}

void dispatcher_add(int fd, void (*event_handler)(msgval *arg))
{
	struct dispatcher_op op;

	op.opcode = EVSOURCE_ADD;
	op.fd = fd;
	op.event_handler = event_handler;

	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));
}

void dispatcher_remove(int fd)
{
	struct dispatcher_op op;

	op.opcode = EVSOURCE_DEL;
	op.fd = fd;

	write(dispatcher_outfd, &op, sizeof(struct dispatcher_op));
}

#endif /* HAVE_SELECT */

#ifndef HAVE_DISPATCHER
#error No event monitoring mechanism is available for this host
#endif /* HAVE_DISPATCHER */
