/**
  @file  ExecQueue.c 
  @brief Lightweight interthread execution control queues
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <pthread.h>
#include <string.h>

#include "ExecQueue.h"
#include "xmalloc.h"

/** To reduce heap usage and fragmentation, we store and recycle message
   queue frames.  This defines how many frames to create initially.  If
   we run out of spare frames we will create more */
#define INITIAL_SPARE_FRAMES	1

/** This designates the maximum number of spare frames to keep on hand */
#define MAX_SPARE_FRAMES	16

/** The message queue is composed of a number of frames which each define
   a message type, a pointer to some data, and a link to the next message
   in the queue */
typedef struct _ExecQueueFrame {
	void (*function)(msgval *);
	msgval arg;

	struct _ExecQueueFrame *link;
} *ExecQueueFrame;

struct _ExecQueue {
	ExecQueueFrame head, tail;
	ExecQueueFrame spare_frames;
	int spare_frame_count;

	pthread_mutex_t lock;
	pthread_cond_t cond;
}; 

ExecQueue execqueue_create()
{
	ExecQueueFrame frame;
	ExecQueue queue = NEW(ExecQueue);

	queue->head = queue->tail = 0;
	queue->spare_frames = 0;

	/* Allocate a reserve of spare frames to keep heap usage down */
	for(queue->spare_frame_count = 0; 
			queue->spare_frame_count < INITIAL_SPARE_FRAMES; 
			queue->spare_frame_count++) 
	{
		frame = NEW(ExecQueueFrame);
		frame->link = queue->spare_frames;
		queue->spare_frames = frame;
	}

	pthread_mutex_init(&queue->lock, 0);
	pthread_cond_init(&queue->cond, 0);

	return queue;
}

void execqueue_destroy(ExecQueue queue)
{
	ExecQueueFrame frame;

	while((frame = queue->spare_frames)) {
		queue->spare_frames = frame->link;
		xfree(frame);
	}

	while((frame = queue->head)) {
		/* XXX Evaluate this loop for any problems which may arise
		   during the termination of an execution queue */
		queue->head = frame->link;
		xfree(frame);
	}

	pthread_cond_destroy(&queue->cond);
	pthread_mutex_destroy(&queue->lock);

	xfree(queue);
}

void execqueue_invoke(ExecQueue queue, void (*function)(msgval *), msgval *arg)
{
	ExecQueueFrame frame;

	pthread_mutex_lock(&queue->lock);

	if(!(frame = queue->spare_frames)) 
		frame = NEW(ExecQueueFrame);
	else {
		queue->spare_frames = frame->link;
		queue->spare_frame_count--;
	}

	if(!queue->head)
		queue->head = queue->tail = frame;
	else {
		queue->tail->link = frame;
		queue->tail = frame;
	}

	frame->function = function;
	if(function) 
		memcpy(&frame->arg, arg, sizeof(msgval));
	
	frame->link = 0;

	pthread_cond_signal(&queue->cond);
	pthread_mutex_unlock(&queue->lock);
}

void execqueue_terminate(ExecQueue queue)
{
	execqueue_invoke(queue, 0, 0);
}

void execqueue_run(ExecQueue queue, void (*postfunc)())
{
	ExecQueueFrame frame;
	void (*function)(msgval *);
	msgval arg;

	for(;;) {
		pthread_mutex_lock(&queue->lock);

		/* If there is no message in the queue then block and 
		   wait for one */
		if(!queue->head)
			pthread_cond_wait(&queue->cond, &queue->lock);

		frame = queue->head;
		queue->head = frame->link;

		function = frame->function;
		if(function) 
			memcpy(&arg, &frame->arg, sizeof(msgval));

		if(queue->spare_frame_count >= MAX_SPARE_FRAMES)
			xfree(frame);
		else {
			frame->link = queue->spare_frames;
			queue->spare_frames = frame;
			queue->spare_frame_count++;
		}

		pthread_mutex_unlock(&queue->lock);

		if(!function) 
			return;
		
		function(&arg);

		if(postfunc)
			postfunc();
	} 
}
