/**
  @file HMAC.c 
  @brief RFC2104 keyed hashing for message authentication
*/
/*
   Copyright (C)2004 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <string.h>

#include "HMAC.h"
#include "Hasher.h"
#include "xmalloc.h"

/** Valid for MD5 and SHA1 according to RFC2104 */
#define BLOCK_SIZE	64

void hmac_compute(HashAlgorithm algorithm, const void *key, unsigned klen, const void *data, unsigned dlen, void *digest)
{
	int i;
	void *kptr;
	uint8_t ipad[BLOCK_SIZE], opad[BLOCK_SIZE];
	Hasher hasher = hasher_create(algorithm);

	if(klen > BLOCK_SIZE) {
		hasher_zero(hasher);
		hasher_add_data(hasher, key, klen);

		klen = hasher_digest_size(hasher);
		kptr = (uint8_t *)xmalloc(klen);
		
		hasher_get_digest(hasher, kptr);
	} else
		kptr = (void *)key;
		
	memset(ipad, 0, BLOCK_SIZE);
	memset(opad, 0, BLOCK_SIZE);
	memcpy(ipad, kptr, klen);
	memcpy(opad, kptr, klen);

	if(key != kptr)
		xfree(kptr);

	for(i = 0; i < BLOCK_SIZE; i++) {
		ipad[i] ^= 0x36;
		opad[i] ^= 0x5c;
	}

	hasher_zero(hasher);
	hasher_add_data(hasher, ipad, BLOCK_SIZE);
	hasher_add_data(hasher, data, dlen);
	hasher_get_digest(hasher, digest);

	hasher_zero(hasher);
	hasher_add_data(hasher, opad, BLOCK_SIZE);
	hasher_add_data(hasher, digest, hasher_digest_size(hasher));
	hasher_get_digest(hasher, digest);

	hasher_destroy(hasher);
}
