#ifndef HASHTABLE_H
#define HASHTABLE_H

/**
 * @file HashTable.h
 * @brief A hash table implementation
 */

/**
 * @defgroup hashtable Hash Table
 * @{
 */

/** HashTable data type */
typedef struct _HashTable *HashTable;

/* Constructor and destructor */

/** @return a handle to a new hash table */
HashTable hash_table_create(void);

/** 
   destroys the given hash table and calls the given function 
   on the pointer associated with each key.  
   @param table the hash table to destroy
   @param * The given function should free the data associated with the keys.  
             If NULL is given instead, no action is performed on the stored pointers, 
             and only internal structures are freed.
*/
void hash_table_destroy(HashTable table, void (*)(void *));

/* Associated operations */

/** 
   Inserts the given zero terminated string key and associated pointer into 
   the table. 
   
  @param table the table
  @param key zero terminated string key
  @param data pointer to insert into the table

  @retval   0	Entry successfully inserted
  @retval  -1	Entry already exists
*/
int hash_table_insert(HashTable table, const char *key, void *data);

/** 
   Removes the entry associated with the given table and removes the associated 
   pointer.

   @retval NULL if no entry is found
   @retval dataptr of the entry if found
*/
void *hash_table_remove(HashTable table, const char *key);

/**
   Returns the pointer associated with the given key, or NULL if not found.
   @retval NULL if not found
   @retval dataptr of the entry if found
*/
void *hash_table_lookup(HashTable table, const char *key);

/** 
 * execute the function on every entry in the table 
 * @param table HashTable to run it on
 * @param iterator the callback function to call, which is passed the data pointer of the entry.
 * @param ptr data structure to pass to the function
 */
void hash_table_iterate(HashTable table, int (*iterator)(void *data, void *ptr), void *ptr);

/** @returns the size of the hash table */
int hash_table_size(HashTable table);

/** @} */
#endif
