/**
   @file Hasher.c 
   @brief Abstract hash cipher interface
*/
/*
   Copyright (C)2004 Anthony Arcieri
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include "Hasher.h"
#include "MD5.h"
#include "SHA1.h"
#include "xmalloc.h"

struct _Hasher {
	unsigned digest_size;
	void *handle;
	
	void (*destroy)(void *);
	void (*zero)(void *);
	void (*add)(void *, const void *, unsigned);
	void (*get_digest)(void *, void *);
};

struct hash_cipher {
	unsigned digest_size;				/**< Bytes in digest  */
	void *(*create)();				/**< Create a hasher */
	void (*destroy)(void *);			/**< Destory a hasher */
	void (*zero)(void *);				/**< Zero hasher */
	void (*add)(void *, const void *, unsigned);	/**< Hash data set */
	void (*get_digest)(void *, void *);		/**< Get digest */
};

const struct hash_cipher ciphers[2] = {
	/** MD5 */
	{ 16, 
	(void *(*)())md5_create, 
	(void (*)(void *))md5_destroy, 
	(void (*)(void *))md5_zero, 
	(void (*)(void *, const void *, unsigned))md5_add_data, 
	(void (*)(void *, void *))md5_get_digest },
	
	/** SHA1 */
	{ 20,
	(void *(*)())sha1_create,
	(void (*)(void *))sha1_destroy,
	(void (*)(void *))sha1_zero,
	(void (*)(void *, const void *, unsigned))sha1_add_data,
	(void (*)(void *, void *))sha1_get_digest }
};

Hasher hasher_create(HashAlgorithm algorithm)
{
	Hasher hasher = NEW(Hasher);

	hasher->digest_size = ciphers[algorithm].digest_size;
	hasher->handle = ciphers[algorithm].create();

	hasher->destroy = ciphers[algorithm].destroy;
	hasher->zero = ciphers[algorithm].zero;
	hasher->add = ciphers[algorithm].add;
	hasher->get_digest = ciphers[algorithm].get_digest;

	return hasher;
}


void hasher_destroy(Hasher hasher)
{
	hasher->destroy(hasher->handle);
	xfree(hasher);
}


unsigned hasher_digest_size(Hasher hasher)
{
	return hasher->digest_size;
}


void hasher_zero(Hasher hasher)
{
	hasher->zero(hasher->handle);
}


void hasher_add_data(Hasher hasher, const void *buf, unsigned len)
{
	hasher->add(hasher->handle, buf, len);
}


void hasher_get_digest(Hasher hasher, void *digest)
{
	hasher->get_digest(hasher->handle, digest);
}
