/**
  @file Lock.c 
  @brief Abstract thread synchronization locks
*/
/*
   Copyright (C)2002-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 

 A quick description of these functions:

 These functions implement a generic interface for two different locking models.
 Exclusive locks act as normal mutexes.  They only allow one thread to hold them
 at a given time.  In this model, there is no difference between the behavior
 of a "reader" thread obtaining the lock and a "writer" thread obtaining it;
 both only allow one thread of any type to acqurire the lock at a given time.

 The shared reader locks allow multiple "reader" threads to hold the lock
 at a given time.  When a "writer" thread requests a lock, all further
 attempts by both "reader" and "writer" threads to acquire the lock will block
 until the writer attempting to obtain the lock has both obtained and
 released it.  The writer will block as it attempts to obtain the lock
 until all reader threads have released the lock.  Thus this model provides
 shared access for readers and exclusive access for writers.
 
 Because of the high overhead of mutexes, this is only advantageous if the code
 section being locked performs some non-trivial functionality (at least 1000
 machine instructions on most architectures), or if a thread has the
 possibility of making a blocking call with the lock held.

 The interface has been designed so that the efficiency of each type of lock
 can be evaluated in a case-by-case basis and easily toggled between models.
 If nothing else the use of separate reader/writer function invocations when
 using the exclusive lock model will help to better document the code.
 */

#include "Lock.h"
#include "threadops.h"
#include "xmalloc.h"

struct _Lock {
	LockType type;
	
	/** Used by both implementations */
	Mutex lock;

	/** Used only by the shared reader lock implementation */
	Mutex rc_lock;
	Cond rc_cond;

	unsigned rc;
};

Lock lock_create(LockType type)
{
	Lock lock = NEW(Lock);

	lock->type = type;
	
	switch(type) {
		case LOCK_SHARED:
			lock->rc = 0;
			lock->rc_cond = thr_cond_create();
			lock->rc_lock = thr_mutex_create();
		case LOCK_EXCL:
			lock->lock = thr_mutex_create();
	}

	return lock;
}

void lock_destroy(Lock lock)
{
	switch(lock->type) {
		case LOCK_SHARED:
			thr_cond_destroy(lock->rc_cond);
			thr_mutex_destroy(lock->rc_lock);
		case LOCK_EXCL:
			thr_mutex_destroy(lock->lock);
	}

	xfree(lock);
}


void lock_reader_acquire(Lock lock)
{
	thr_mutex_lock(lock->lock);

	if(lock->type == LOCK_EXCL)
		return;

	thr_mutex_lock(lock->rc_lock);
	lock->rc++;
	thr_mutex_unlock(lock->rc_lock);

	thr_mutex_unlock(lock->lock);
}

void lock_reader_release(Lock lock)
{
	if(lock->type == LOCK_EXCL) {
		thr_mutex_unlock(lock->lock);
		return;
	}

	thr_mutex_lock(lock->rc_lock);
	
	if(!--lock->rc)
		thr_cond_signal(lock->rc_cond);

	thr_mutex_unlock(lock->rc_lock);
}


void lock_writer_acquire(Lock lock)
{
	thr_mutex_lock(lock->lock);

	if(lock->type == LOCK_EXCL)
		return;

	thr_mutex_lock(lock->rc_lock);

	if(lock->rc > 0)
		thr_cond_wait(lock->rc_cond, lock->rc_lock);
}

void lock_writer_release(Lock lock)
{
	if(lock->type == LOCK_SHARED)
		thr_mutex_unlock(lock->rc_lock);

	thr_mutex_unlock(lock->lock);
}
