/**
  @file MsgQueue.c 
  @brief Lightweight interthread message queues
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <string.h>

#include "MsgQueue.h"
#include "msgval.h"
#include "threadops.h"
#include "xmalloc.h"

/** To reduce heap usage and fragmentation, we store and recycle message
   queue frames.  This defines how many frames to create initially.  If
   we run out of spare frames we will create more */
#define INITIAL_SPARE_FRAMES	1

/** This designates the maximum number of spare frames to keep on hand */
#define MAX_SPARE_FRAMES	16

/** The message queue is composed of a number of frames which each define
   a message type, a pointer to some data, and a link to the next message
   in the queue */
typedef struct _MsgQueueFrame {
	msgval arg;
	struct _MsgQueueFrame *link;
} *MsgQueueFrame;

struct _MsgQueue {
	MsgQueueFrame head, tail;
	MsgQueueFrame spare_frames;
	int spare_frame_count;

	Mutex lock;
	Cond cond;
}; 

MsgQueue msgqueue_create()
{
	MsgQueueFrame frame;
	MsgQueue queue = NEW(MsgQueue);

	queue->head = queue->tail = 0;
	queue->spare_frames = 0;

	/* Allocate a reserve of spare frames to keep heap usage down */
	for(queue->spare_frame_count = 0; 
			queue->spare_frame_count < INITIAL_SPARE_FRAMES; 
			queue->spare_frame_count++) 
	{
		frame = NEW(MsgQueueFrame);
		frame->link = queue->spare_frames;
		queue->spare_frames = frame;
	}

	queue->lock = thr_mutex_create();
	queue->cond = thr_cond_create();

	return queue;
}

void msgqueue_destroy(MsgQueue queue)
{
	MsgQueueFrame frame;

	while((frame = queue->spare_frames)) {
		queue->spare_frames = frame->link;
		xfree(frame);
	}

	while((frame = queue->head)) {
		/* XXX Evaluate this loop for any problems which may arise
		   during the termination of an execution queue */
		queue->head = frame->link;
		xfree(frame);
	}

	thr_cond_destroy(queue->cond);
	thr_mutex_destroy(queue->lock);

	xfree(queue);
}

void msgqueue_send(MsgQueue queue, msgval *arg)
{
	MsgQueueFrame frame;

	thr_mutex_lock(queue->lock);

	if(!(frame = queue->spare_frames)) 
		frame = NEW(MsgQueueFrame);
	else {
		queue->spare_frames = frame->link;
		queue->spare_frame_count--;
	}

	if(!queue->head)
		queue->head = queue->tail = frame;
	else {
		queue->tail->link = frame;
		queue->tail = frame;
	}

	memcpy(&frame->arg, arg, sizeof(msgval));

	frame->link = 0;

	thr_cond_signal(queue->cond);
	thr_mutex_unlock(queue->lock);
}

void msgqueue_recv(MsgQueue queue, msgval *arg)
{
	MsgQueueFrame frame;

	thr_mutex_lock(queue->lock);

	/* If there is no message in the queue then block and 
	   wait for one */
	if(!queue->head)
		thr_cond_wait(queue->cond, queue->lock);

	frame = queue->head;
	queue->head = frame->link;

	memcpy(arg, &frame->arg, sizeof(msgval));

	if(queue->spare_frame_count >= MAX_SPARE_FRAMES)
		xfree(frame);
	else {
		frame->link = queue->spare_frames;
		queue->spare_frames = frame;
		queue->spare_frame_count++;
	}

	thr_mutex_unlock(queue->lock);
}
