/**
  @file RNG.c 
  @brief Random number generation routines
*/
/*
   Copyright (C)2003-04 Anthony Arcieri, Matt Bishop
   All rights reserved.

   For licensing information, please see the file LICENSE 
 */

#include <machdep.h>

#ifdef WIN32
#include <wincrypt.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#endif

#include "Config.h"
#include "RNG.h"
#include "log.h"
#include "xmalloc.h"

#ifdef WIN32
static HCRYPTPROV cp;
#else

/** Default entropy gathering device */
#define ENTROPY_DEVICE	"/dev/urandom"

static int random_device_fd = -1;
#endif

/** Initialize random number generator */
void rng_init()
{
#ifdef WIN32
	if(!CryptAcquireContext(&cp, 0, MS_STRONG_PROV, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) {
		log_error("CryptAcquireContext");
		fatal("!!! Couldn't initialize random number generator");
	}
#else
	char *random_dev;

	if(random_device_fd != -1)
		return;

	if(!(random_dev = config_value("global", "random_device"))) {
		log_printf("*** global::random_device unset, defaulting to %s", ENTROPY_DEVICE);
		random_dev = xstrdup(ENTROPY_DEVICE);
	}

	if((random_device_fd = open(random_dev, O_RDONLY)) < 0) {
		log_error(random_dev);
		fatal("!!! Couldn't initialize random number generator");
	}

	xfree(random_dev);
#endif
}

/** Shut down random number generator */
void rng_destroy()
{
#ifdef WIN32
	CryptReleaseContext(cp, 0);
#else
	close(random_device_fd);
#endif
}

/** Generate a random buffer of the given length */
void rng_randbuf(void *buffer, unsigned len)
{
#ifdef WIN32
	if(!CryptGenRandom(cp, len, buffer)) {
		log_error("CryptGenRandom");
		fatal("!!! Random number generator error");
	}
#else
	if(read(random_device_fd, buffer, len) < 0) {
		log_error("read");
		fatal("!!! Random number generator error");
	}
#endif
}
