/**
  @file SysDir.c
  @brief Abstract system directory listing interface
 */
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

/* System headers */
#include <string.h>
#ifndef WIN32
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#endif

#include "DirEntry.h"
#include "SysDir.h"
#include "debug.h"
#include "util.h"
#include "xmalloc.h"

struct _SysDir {
#ifdef WIN32
	HANDLE find;
	WIN32_FIND_DATA data;
#else
	DIR *dirp;
	char *path;
#endif
};

SysDir sysdir_open(const char *path)
{
	SysDir sysdir;

#ifdef WIN32
	HANDLE find;
	WIN32_FIND_DATA data;
	char *syspath, findpath[PDTP_PATH_MAX];
#else
	DIR *dirp;
#endif /* WIN32 */

#ifdef WIN32
	syspath = xstrdup(path);
	add_trailing_slash(&syspath);
	safecat(findpath, PDTP_PATH_MAX, syspath, "*", 0);
	xfree(syspath);

	if((find = FindFirstFile(findpath, &data)) == INVALID_HANDLE_VALUE) {
#ifndef NDEBUG
		debug("Opening of \'%s\' failed", findpath);
#endif
		return 0;
	}

#else
	if(!(dirp = opendir(path)))
		return 0;
#endif /* WIN32 */

	sysdir = NEW(SysDir);

#ifdef WIN32
	sysdir->find = find;
	memcpy(&sysdir->data, &data, sizeof(WIN32_FIND_DATA));
#else
	sysdir->dirp = dirp;
	sysdir->path = xstrdup(path);
#endif /* WIN32 */

	return sysdir;
}

void sysdir_close(SysDir sysdir)
{
#ifdef WIN32
	if(sysdir->find)
		FindClose(sysdir->find);
#else
	closedir(sysdir->dirp);
	xfree(sysdir->path);
#endif /* WIN32 */

	xfree(sysdir);
}

int sysdir_get_entry(SysDir sysdir, struct DirEntry *entry)
{
#ifdef WIN32
	LARGE_INTEGER fsize, ctime, yeartime;
	SYSTEMTIME systime;
	FILETIME ftime;
#else
	int r = -1, fd;
	struct dirent *dentry;
	struct stat st;
	struct tm tm;
	char path[PATH_MAX];
#endif /* WIN32 */

#ifdef WIN32
	if(!sysdir->find)
		return -1;

	if(!FindNextFile(sysdir->find, &sysdir->data)) {
		FindClose(sysdir->find);
		sysdir->find = 0;
	}

	if(sysdir->data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
		entry->type = ENTRY_DIR;
		entry->length = 0;
	} else {
		entry->type = ENTRY_FILE;

		fsize.HighPart = sysdir->data.nFileSizeHigh;
		fsize.LowPart = sysdir->data.nFileSizeLow;
		entry->length = fsize.QuadPart;
	}

	safecpy(entry->name, sysdir->data.cFileName, PDTP_PATH_MAX);

	FileTimeToSystemTime(&sysdir->data.ftLastWriteTime, &systime);

	entry->ts_year = systime.wYear;

	/* Set values relative to January 1st */
	systime.wMonth = 1;
	systime.wDay = 1;
	systime.wHour = 0;
	systime.wMinute = 0;
	systime.wSecond = 0;
	systime.wMilliseconds = 0;

	SystemTimeToFileTime(&systime, &ftime);

	yeartime.HighPart = ftime.dwHighDateTime;
	yeartime.LowPart = ftime.dwLowDateTime;

	ctime.HighPart = sysdir->data.ftLastWriteTime.dwHighDateTime;
	ctime.LowPart = sysdir->data.ftLastWriteTime.dwLowDateTime;

	entry->ts_ysec = (int)((ctime.QuadPart - yeartime.QuadPart) / 10000000);
#else
	/* Continue scanning even if we hit files we can't read */
	do {
		if(!(dentry = readdir(sysdir->dirp)))
			return -1;

		safecat(path, PATH_MAX, sysdir->path, "/", dentry->d_name, 0);
		
		if((fd = open(path, O_RDONLY)) < 0)
			continue;

		r = fstat(fd, &st);
		close(fd);
	} while(r < 0);

	if(S_ISDIR(st.st_mode)) {
		entry->type = ENTRY_DIR;
		entry->length = 0;
	} else {
		entry->type = ENTRY_FILE;
		entry->length = st.st_size;
	}

	safecpy(entry->name, dentry->d_name, PDTP_PATH_MAX);

	localtime_r(&st.st_mtime, &tm);
	entry->ts_year = tm.tm_year + 1900;
	entry->ts_ysec = tm.tm_yday * 86400 + tm.tm_hour * 3600 + tm.tm_min * 60 + tm.tm_sec;
#endif /* WIN32 */

	return 0;
}
