/**
  @file Transaction.c 
  @brief Code for constructing and processing transactions
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <string.h>
#include <time.h>
#include <stdarg.h>

#include "Transaction.h"
#include "socketops.h"
#include "xmalloc.h"
#include "debug.h"

#ifndef NDEBUG
#include <stdio.h>
#include <ctype.h>
#endif

/* Include uio.h if we are using UIO */
#ifdef HAVE_UIO
#include <sys/uio.h>
#endif

/** Sanity limit for the size of a transaction (2MB) */
#define	MAX_TXN_SIZE	0x200000

/* Transaction type bit */
#define TXN_REPLY_BIT	0x8000

struct TransactionObject {
	uint32_t obj_len;
	uint16_t obj_id;
	void *obj_data;
};

struct _TransactionIn {
	TransactionType type;
	uint32_t serial;
	uint16_t id;

	uint8_t *buffer;
	struct TransactionObject *object;
	uint16_t current_obj, obj_count;
};

struct _TransactionOut {
	uint32_t length;
	uint32_t serial;
	uint16_t id;

	uint16_t obj_count, max_objects;

#ifdef HAVE_UIO
	struct iovec *iov;
#endif

#ifdef WIN32
	LPWSABUF buf;
#endif

#if !defined(HAVE_UIO) && !defined(WIN32)
	struct TransactionObject *object;
#endif
};

/**
   XXX Read an incoming transaction and create a transaction object.  Zero 
   copy, but doesn't use persistent buffering.  This may lead to some nasty
   heap usage patterns, namely fragmentation.  Persistent buffering for small
   transactions would help this dramatically.
 */
TransactionIn transaction_read(int sock)
{
	TransactionIn txn;
	uint8_t *bptr;
	uint16_t max_objects;
	uint32_t length;

	uint16_t v16;
	uint32_t v32;

	uint8_t header[12];

	if(readsock(sock, header, 12) < 0) {
#ifndef NDEBUG
		debug("*** TXN_ERR: read() error while reading header");
#endif
		return 0;
	}

	memcpy(&v32, header, 4);
	length = ntohl(v32);

	if(length > MAX_TXN_SIZE) {
#ifndef NDEBUG
		debug("*** TXN_ERR: Transaction too long (%d)", length);
#endif
		return 0;
	}

	/* Subtract the length of the header */
	length -= 8;

	txn = NEW(TransactionIn);

	txn->current_obj = 0;

	memcpy(&v32, header + 4, 4);
	txn->serial = ntohl(v32);

	memcpy(&v16, header + 8, 2);
	txn->id = ntohs(v16);

	if(txn->id & TXN_REPLY_BIT) {
		txn->type = TXN_REPLY;
		txn->id &= ~TXN_REPLY_BIT;
	} else
		txn->type = TXN_COMMAND;

	memcpy(&v16, header + 10, 2);
	max_objects = ntohs(v16);

#ifndef NDEBUG
	debug("--- | INCOMING TRANSACTION | ---");
	debug("Type:	%s", txn->type == TXN_COMMAND ? "command" : "reply");
	debug("ID:	0x%x", txn->id);
	debug("Length:  %d", length);
	debug("Objects: %d", max_objects);
#endif

	if(!max_objects) {
		if(length > 0)
			goto err0;

		txn->obj_count = 0;

		return txn;
	}

	/* XXX Here is where we abuse the heap.  This should be replaced with 
	   a reasonably sized fixed length buffer. */

	txn->buffer = (uint8_t *)xmalloc(length);
	if(readsock(sock, txn->buffer, length) < 0) {
#ifndef NDEBUG
		debug("*** TXN_ERR: read() error reading body");
#endif
		goto err0;
	}

	txn->object = (struct TransactionObject *)
		xmalloc(max_objects * sizeof(struct TransactionObject));

	for(txn->obj_count = 0, bptr = txn->buffer; txn->obj_count < max_objects; txn->obj_count++) {
		if(length < 6) {
#ifndef NDEBUG
			debug("*** TXN_ERR: Length mismatch in object header");
#endif
			goto err1;
		}

		memcpy(&v32, bptr, 4);
		txn->object[txn->obj_count].obj_len = ntohl(v32);

		bptr += 4;
		length -= 4;

		memcpy(&v16, bptr, 2);
		txn->object[txn->obj_count].obj_id = ntohs(v16);

		bptr += 2;
		length -= 2;

		if(length < txn->object[txn->obj_count].obj_len) {
#ifndef NDEBUG
			debug("*** TXN_ERR: Length mismatch in object body");
#endif
			goto err1;
		}

#ifndef NDEBUG
		debug("Object length is %d", txn->object[txn->obj_count].obj_len);
#endif
		txn->object[txn->obj_count].obj_data = bptr;

		bptr += txn->object[txn->obj_count].obj_len;
		length -= txn->object[txn->obj_count].obj_len;
	}

	if(length) {
#ifndef NDEBUG
		debug("*** TXN_ERR: Garbage on end of transaction");
#endif
		goto err1;
	}

#ifndef NDEBUG
	debug("*** Got transaction: %p", txn);
#endif

	return txn;

err1:
	xfree(txn->object);
err0:
	xfree(txn->buffer);
	xfree(txn);

	return 0;
}

void transaction_in_destroy(TransactionIn txn)
{
	if(txn->obj_count) {
		xfree(txn->object); 
		xfree(txn->buffer);
	}

	xfree(txn);
}

TransactionType transaction_type(TransactionIn txn)
{
	return txn->type;
}

uint16_t transaction_id(TransactionIn txn)
{
	return txn->id;
}

uint32_t transaction_serial(TransactionIn txn)
{
	return txn->serial;
}

int transaction_object_count(TransactionIn txn)
{
	return txn->obj_count;
}

int transaction_object_index(TransactionIn txn, uint16_t id)
{
        while(txn->current_obj < txn->obj_count) {
                if(txn->object[txn->current_obj].obj_id >= id)
                        break;

                txn->current_obj++;
        }

        if(txn->current_obj == txn->obj_count || txn->object[txn->current_obj].obj_id != id)
                return -1;

        return txn->current_obj;
}

uint16_t transaction_object_id(TransactionIn txn, unsigned objno)
{
	return txn->object[objno].obj_id;
}

int transaction_object_length(TransactionIn txn, unsigned objno)
{
	return txn->object[objno].obj_len;
}

void *transaction_object_data(TransactionIn txn, unsigned objno)
{
	return txn->object[objno].obj_data;
}

int transaction_object_uint16(TransactionIn txn, unsigned objno, uint16_t *object)
{
	uint16_t value;
	
	if(objno >= txn->obj_count)
		return -1;
	
	if(txn->object[objno].obj_len != 2)
		return -1;

	memcpy(&value, txn->object[objno].obj_data, 2);
	*object = ntohs(value);

	return 0;
}

int transaction_object_uint32(TransactionIn txn, unsigned objno, uint32_t *object)
{
	uint32_t value;

	if(objno >= txn->obj_count)
		return -1;

	if(txn->object[objno].obj_len != 4)
		return -1;

	memcpy(&value, txn->object[objno].obj_data, 4);
	*object = ntohl(value);

	return 0;
}

char *transaction_object_string(TransactionIn txn, unsigned objno)
{
	char *ret = (char *)xmalloc(txn->object[objno].obj_len + 1);
	memcpy(ret, txn->object[objno].obj_data, txn->object[objno].obj_len);
	ret[txn->object[objno].obj_len] = '\0';

	return ret;
}

int transaction_object_unpack(TransactionIn txn, unsigned objno, const char *format, ...)
{
#ifndef HOST_BIGENDIAN
	int i;
#endif
	int ret = 0, length = txn->object[objno].obj_len;
	uint8_t *bptr = (uint8_t *)txn->object[objno].obj_data;
	uint16_t v16;
	uint32_t v32;
	uint64_t v64;
	va_list ap;
	char *str;

	va_start(ap, format);

	while(*format) {
		switch(*format) {
			case '1':
				if(length < 1)
					goto done;

				*va_arg(ap, int *) = *bptr;

				length -= 1;
				bptr += 1;

				break;
			case '2':
				if(length < 2)
					goto done;

				memcpy(&v16, bptr, 2);
				*va_arg(ap, int *) = htons(v16);

				length -= 2;
				bptr += 2;

				break;
			case '4':
				if(length < 4)
					goto done;

				memcpy(&v32, bptr, 4);
				*va_arg(ap, uint32_t *) = htonl(v32);

				length -= 4;
				bptr += 4;

				break;
			case '8':
				if(length < 8)
					goto done;

#ifdef HOST_BIGENDIAN
				memcpy(&v64, bptr, 8);
#else
				for(i = 0; i < 8; i++)
					((uint8_t *)&v64)[i] = bptr[7 - i];
#endif
				*va_arg(ap, uint64_t *) = v64;

				length -= 8;
				bptr += 8;

				break;
			case 's':
				if(length < 2)
					goto done;

				/* Extract the string length into v16 */
				memcpy(&v16, bptr, 2);
				v16 = htons(v16);

				length -= 2;
				bptr += 2;

				if(length < v16)
					goto done;

				str = (char *)xmalloc(v16 + 1);
				memcpy(str, bptr, v16);
				str[v16] = '\0';
				*va_arg(ap, char **) = str;

				length -= v16;
				bptr += v16;
		}

		format++;
		ret++;
	}

done:
	va_end(ap);
	return ret;
}

TransactionOut transaction_create(TransactionType type, uint16_t tid, uint16_t obj_count)
{
	uint16_t v16;
	
	TransactionOut txn = NEW(TransactionOut);

	txn->serial = 0;

	v16 = tid;
	if(type == TXN_REPLY)
		v16 |= TXN_REPLY_BIT;
	
	txn->id = htons(v16);
	txn->length = 8;
	txn->obj_count = 0;
	txn->max_objects = obj_count;

#ifdef HAVE_UIO
	txn->iov = (struct iovec *)xmalloc((4 + txn->max_objects) * sizeof(struct iovec));

	/* This iov_base field gets filled in before the writev call */
	txn->iov[0].iov_len = 4;

	txn->iov[1].iov_len = 4;
	txn->iov[1].iov_base = (char *)&txn->serial;

	txn->iov[2].iov_len = 2;
	txn->iov[2].iov_base = (char *)&txn->id;

	/* This iov_base field gets filled in before the writev call */
	txn->iov[3].iov_len = 2;
#endif

#ifdef WIN32
        txn->buf = (LPWSABUF)xmalloc((4 + txn->max_objects) * sizeof(WSABUF));

        /* This buf field gets filled in before the WSASend call */
        txn->buf[0].len = 4;

        txn->buf[1].len = 4;
        txn->buf[1].buf = (char *)&txn->serial;

        txn->buf[2].len = 2;
        txn->buf[2].buf = (char *)&txn->id;

        /* This buf field gets filled in before the WSASend call */
        txn->buf[3].len = 2;
#endif

#if !defined(HAVE_UIO) && !defined(WIN32)
	txn->object = (struct TransactionObject *)xmalloc(txn->max_objects * sizeof(struct TransactionObject));
#endif

	return txn;
}

TransactionOut transaction_copy(TransactionIn txnin)
{
	int i;
	TransactionOut txnout;

	txnout = transaction_create(txnin->type, txnin->id, txnin->obj_count);
	
	for(i = 0; i < txnin->obj_count; i++)
		transaction_add_object(txnout, txnin->object[i].obj_id, txnin->object[i].obj_data, txnin->object[i].obj_len);

	return txnout;
}

void transaction_set_serial(TransactionOut txn, uint32_t serial)
{
	txn->serial = htonl(serial);
}

void transaction_out_destroy(TransactionOut txn)
{
	int i;

#ifdef HAVE_UIO
	for(i = 0; i < txn->obj_count; i++)
		xfree(txn->iov[4 + i].iov_base);

	xfree(txn->iov);
#endif

#ifdef WIN32
        for(i = 0; i < txn->obj_count; i++)
                xfree(txn->buf[4 + i].buf);

        xfree(txn->buf);
#endif

#if !defined(HAVE_UIO) && !defined(WIN32)
	for(i = 0; i < txn->obj_count; i++)
		xfree(txn->object[i].obj_data);

	xfree(txn->object);
#endif

	xfree(txn);
}

void transaction_add_object(TransactionOut txn, uint16_t id, const void *data, uint32_t length)
{
#if defined(HAVE_UIO) || defined(WIN32)
	uint16_t v16;
	uint32_t v32;
	uint8_t *obj_buf;

	if(txn->obj_count >= txn->max_objects)
		return;

	obj_buf = (uint8_t *)xmalloc(6 + length);

	v32 = htonl(length);
	memcpy(obj_buf, &v32, 4);
	v16 = htons(id);
	memcpy(obj_buf + 4, &v16, 2);

	if(length > 0) 
		memcpy(obj_buf + 6, data, length);

#ifdef HAVE_UIO
	txn->iov[txn->obj_count + 4].iov_len = 6 + length;
	txn->iov[txn->obj_count + 4].iov_base = (char *)obj_buf;
#endif /* HAVE_UIO */

#ifdef WIN32
        txn->buf[txn->obj_count + 4].len = 6 + length;
        txn->buf[txn->obj_count + 4].buf = (char *)obj_buf;
#endif

#else
	if(txn->obj_count >= txn->max_objects) 
		return;

	txn->object[txn->obj_count].obj_len = length;
	txn->object[txn->obj_count].obj_id = id;
	txn->object[txn->obj_count].obj_data = xmalloc(length);
	memcpy(txn->object[txn->obj_count].obj_data, data, length);
#endif

	txn->obj_count++;
	txn->length += 6 + length;
}

void transaction_add_int16(TransactionOut txn, uint16_t id, uint16_t value)
{
	uint16_t v = htons(value);
	transaction_add_object(txn, id, &v, 2);
}

void transaction_add_int32(TransactionOut txn, uint16_t id, uint32_t value)
{
	uint32_t v = htonl(value);
	transaction_add_object(txn, id, &v, 4);
}

void transaction_add_int64(TransactionOut txn, uint16_t id, uint64_t value)
{
#ifdef HOST_BIGENDIAN
	transaction_add_object(txn, id, &value, 8);
#else 
	int i;
	uint8_t buf[8];

	for(i = 0; i < 8; i++)
		buf[i] = ((uint8_t *)&value)[7 - i];

	transaction_add_object(txn, id, buf, 8);
#endif /* HOST_BIGENDIAN */
}

void transaction_add_string(TransactionOut txn, uint16_t id, const char *string)
{
	transaction_add_object(txn, id, string, (uint32_t)strlen(string));
}

/** Format flags for pack/unpack:

1: 8-bit integer
2: 16-bit integer
4: 32-bit integer
8: 64-bit integer
s: Arbitrary length string

 */

int transaction_object_pack(TransactionOut txn, uint16_t id, const char *format, ...)
{
	typedef struct _PackNode {
		int length;
		uint8_t *buf;

		struct _PackNode *link;
	} *PackNode;

	PackNode head = 0, tail = 0, node;

	unsigned l, length = 0;
	void *buf;
	uint8_t *bptr;
	uint16_t v16;
	uint32_t v32;
	uint64_t v64;
	va_list ap;

#ifndef HOST_BIGENDIAN
	int i;
#endif

	va_start(ap, format);

	while(*format) {
		node = NEW(PackNode);
		node->link = 0;

		switch(*format) {
			case '1':
				length += 1;
				node->length = 1;
				node->buf = (uint8_t *)xmalloc(1);
				*node->buf = va_arg(ap, int);
				break;
			case '2':
				length += 2;
				node->length = 2;
				node->buf = (uint8_t *)xmalloc(2);
				v16 = htons((uint16_t)va_arg(ap, int));
				memcpy(node->buf, &v16, 2);
				break;
			case '4':
				length += 4;
				node->length = 4;
				node->buf = (uint8_t *)xmalloc(4);
				v32 = htonl(va_arg(ap, uint32_t));
				memcpy(node->buf, &v32, 4);
				break;
			case '8':
				length += 8;
				node->length = 8;
				node->buf = (uint8_t *)xmalloc(8);
				v64 = va_arg(ap, uint64_t);
#ifdef HOST_BIGENDIAN
				memcpy(node->buf, &v64, 8);
#else
				for(i = 0; i < 8; i++)
					node->buf[7 - i] = ((uint8_t *)&v64)[i];
#endif
				break;
			case 's':
				if(!(bptr = (uint8_t *)va_arg(ap, char *)))
					bptr = (uint8_t *)"";

				l = (unsigned)strlen((const char *)bptr);
				length += l + 2;
				node->length = l + 2;
				node->buf = (uint8_t *)xmalloc(l + 2);
				v16 = htons((uint16_t)l);
				memcpy(node->buf, &v16, 2);
				memcpy(node->buf + 2, bptr, l);
				break;
#ifndef NDEBUG
			default:
				debug("Warning: Unknown pack code: '%c'", *format);
#endif
		}

		if(!head) 
			head = tail = node;
		else {
			tail->link = node;
			tail = tail->link;
		}

		format++;
	}

	buf = xmalloc(length);
	bptr = (uint8_t *)buf;

	while((node = head)) {
		head = node->link;

		memcpy(bptr, node->buf, node->length);
		bptr += node->length;

		xfree(node->buf);
		xfree(node);
	}

	transaction_add_object(txn, id, buf, length);
	xfree(buf);
	va_end(ap);

	return length;
}

int transaction_write(int sock, TransactionOut txn)
{
	int ret;

#ifdef HAVE_UIO
	uint32_t length;
	uint16_t obj_count;

	length = htonl(txn->length);
	obj_count = htons(txn->obj_count);

	txn->iov[0].iov_base = (char *)&length;
	txn->iov[3].iov_base = (char *)&obj_count;

	ret = writev(sock, txn->iov, 4 + txn->obj_count);
#endif

#ifdef WIN32
        uint32_t length;
        uint16_t obj_count;
        DWORD bytes = 0;

        length = htonl(txn->length);
        obj_count = htons(txn->obj_count);

        txn->buf[0].buf = (char *)&length;
        txn->buf[3].buf = (char *)&obj_count;

	if(WSASend(sock, txn->buf, 4 + txn->obj_count, &bytes, 0, 0, 0) < 0) 
		ret = -1;
	else
		ret = bytes;
#endif

#if !defined(HAVE_UIO) && !defined(WIN32)
	int i;
	uint8_t *buffer, *bptr;
	uint16_t v16;
	uint32_t v32;

	bptr = buffer = (uint8_t *)xmalloc(txn->length + 4);

	v32 = htonl(txn->length);
	memcpy(buffer, &v32, 4);

	memcpy(buffer + 4, &txn->serial, 4);
	memcpy(buffer + 8, &txn->id, 2);

	v16 = htons(txn->obj_count);
	memcpy(buffer + 10, &v16, 2);

	bptr += 12;

	for(i = 0; i < txn->obj_count; i++) {
		v32 = htonl(txn->object[i].obj_len);
		memcpy(bptr, &v32, 4);

		v16 = htons(txn->object[i].obj_id);
		memcpy(bptr + 4, &v16, 2);

		memcpy(bptr + 6, txn->object[i].obj_data, txn->object[i].obj_len);

		bptr += 6 + txn->object[i].obj_len;
	}

	ret = send(sock, buffer, txn->length + 4, 0);
	xfree(buffer);
#endif /* HAVE_UIO */

#ifndef NDEBUG
        debug("--- | OUTGOING TRANSACTION | ---");
        debug("id:       %x", ntohs(txn->id));
        debug("serial:   %d", ntohl(txn->serial));
        debug("length:   %d", txn->length);
        debug("objcnt:   %d", txn->obj_count);
#endif

	return ret;
}
