/**
  @file WorkerManager.c 
  @brief Regulates and maintains thread pool
*/
/*
   Copyright (C)2002-03 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <pthread.h>
#include <stdlib.h>

#include "Config.h"
#include "Stack.h"
#include "ExecQueue.h"
#include "Worker.h"
#include "WorkerManager.h"
#include "debug.h"
#include "log.h"
#include "xmalloc.h"

static Stack inactive_threads = 0;
static int active_thread_count = 0;
static int max_active_threads;
static int max_inactive_threads;

static pthread_mutex_t workermgr_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t workermgr_cond = PTHREAD_COND_INITIALIZER;

void workermgr_init(void)
{
	int i, initial_threads = -1;
	Worker t;

	pthread_mutex_lock(&workermgr_mutex);

	if(inactive_threads == 0)
		inactive_threads = stack_create();

	if((max_inactive_threads = config_int_value("threads", "spare_threads")) < 1) {
		log_printf("*** threads::spare_threads invalid or unspecified, defaulting to %d", DEFAULT_MAX_INACTIVE_THREADS);
		max_inactive_threads = DEFAULT_MAX_INACTIVE_THREADS;
	}

	if((max_active_threads = config_int_value("threads", "pool_size")) < 1) {
		log_printf("*** threads::pool_size invalid or unspecified, defaulting to %d", DEFAULT_MAX_ACTIVE_THREADS);
		max_active_threads = DEFAULT_MAX_ACTIVE_THREADS;
	}

	if(initial_threads < 0) {
		initial_threads = DEFAULT_INITIAL_THREADS;
	}

	for(i = stack_size(inactive_threads) + active_thread_count; i < initial_threads; i++) {
		if(!(t = worker_create()))
			log_printf("!!! Warning: Error spawning thread while populating thread pool");
		else
			stack_push(inactive_threads, t);
	}

	pthread_mutex_unlock(&workermgr_mutex);
}

void workermgr_destroy(void)
{
	Worker worker;
	
	/* Begin by waiting for all active worker threads to complete */
	pthread_mutex_lock(&workermgr_mutex);

	while(active_thread_count > 0) 
		pthread_cond_wait(&workermgr_cond, &workermgr_mutex);
	
	pthread_mutex_unlock(&workermgr_mutex);

	/* Now signal all workers to exit and wait for them to do so */
	while((worker = (Worker)stack_pop(inactive_threads))) 
		worker_destroy(worker);

	/* Free the inactive threads stack structure */
	stack_destroy(inactive_threads);

	/* Free mutexes and conditions */
	pthread_mutex_destroy(&workermgr_mutex);
	pthread_cond_destroy(&workermgr_cond);
}

void workermgr_exec(void (*function)(msgval *), msgval *arg)
{
	Worker worker;

	pthread_mutex_lock(&workermgr_mutex);

#ifndef NDEBUG
	debug("Requesting thread (Active: %d, Inactive: %d)", active_thread_count, stack_size(inactive_threads));
#endif

	while(active_thread_count >= max_active_threads)
		pthread_cond_wait(&workermgr_cond, &workermgr_mutex);

	if(!(worker = (Worker)stack_pop(inactive_threads)) && 
			(worker = worker_create()) == 0) 
	{
		log_printf("!!! Warning: threads::pool_size exceeds maximum number of threads supported by system");
		pthread_cond_wait(&workermgr_cond, &workermgr_mutex);
		if((worker = (Worker)stack_pop(inactive_threads)) == 0) 
			fatal("!!! Fatal error: Thread pool depleted - system may be low on resources");
	}

	active_thread_count++;

	worker_exec(worker, function, arg);
	
	pthread_mutex_unlock(&workermgr_mutex);
}

void workermgr_return_worker(Worker t)
{
	pthread_mutex_lock(&workermgr_mutex);

	if(stack_size(inactive_threads) >= max_inactive_threads)
		worker_destroy(t);
	else 
		stack_push(inactive_threads, t);

	active_thread_count--;

#ifndef NDEBUG
	debug("Thread returned (Active: %d, Inactive: %d)", active_thread_count, stack_size(inactive_threads));
#endif

	pthread_cond_signal(&workermgr_cond);
	pthread_mutex_unlock(&workermgr_mutex);
}

void workermgr_handle_crash(void)
{
	pthread_mutex_lock(&workermgr_mutex);

	active_thread_count--;
	
#ifndef NDEBUG
	debug("Thread crashed (Active: %d, Inactive: %d)\n", active_thread_count, stack_size(inactive_threads));
#endif

	pthread_cond_signal(&workermgr_cond);
	pthread_mutex_unlock(&workermgr_mutex);
}
