/**
  @file log.c 
  @brief Logging functions
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#ifndef WIN32
#include <errno.h>
#endif

#include "log.h"
#include "util.h"
#include "threadops.h"

void log_init()
{
	log_printf("--- | Logfile started | ---");
}

void log_shutdown()
{
	log_printf("--- | Logfile ended | ---");
}

void log_printf(const char *format, ...)
{
	va_list ap;
	struct systime st;
	static Mutex log_mutex = 0;

	if(!log_mutex)
		log_mutex = thr_mutex_create();

	thr_mutex_lock(log_mutex);
	get_systime(&st);

	fprintf(stderr, "[%02d:%02d:%02d] ", st.hour, st.minute, st.second);

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fputc('\n', stderr);
	thr_mutex_unlock(log_mutex);
}

void log_error(const char *s)
{
#ifdef WIN32
	char *msgbuf, *t;

	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 0, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &msgbuf, 0, 0);
	
	if((t = strpbrk(msgbuf, "\r\n")))
		*t = '\0';
	
	log_printf("*** Error: %s: %s", s, msgbuf);

	LocalFree(msgbuf);
#endif /* WIN32 */
	
	/* For platforms without a thread-safe strerror() */
#ifdef HAVE_STRERROR_R
	char msgbuf[256] = { 0 };

	strerror_r(errno, msgbuf, 256);
	log_printf("*** Error: %s: %s", s, msgbuf);
#endif

#if !defined(WIN32) && !defined(HAVE_STRERROR_R)
	log_printf("*** Error: %s: %s", s, strerror(errno));
#endif
}

void fatal(const char *format, ...)
{
	va_list ap;

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fputc('\n', stderr);

	exit(EXIT_FAILURE);
}
