/**
  @file socketops.c 
  @brief Platform independent socket functions
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <stdio.h>
#include <string.h>
#ifndef WIN32
#include <errno.h>
#endif

#include "global.h"
#include "socketops.h"
#include "xmalloc.h"

/** Timeout for connect() attempts */
#define CONNECT_TIMEOUT         30

/** Maximum backlog for listener sockets */
#define MAX_BACKLOG		16

struct _Addr {
	struct sockaddr *saddr;
	socklen_t slen;
};

Addr addr_create_from_ipv4addr(uint32_t addr, uint16_t port)
{
	Addr ret = NEW(Addr);
	struct sockaddr_in *sin;

	ret->saddr = (struct sockaddr *)xmalloc(sizeof(struct sockaddr_in));
	ret->slen = sizeof(struct sockaddr_in);
	memset(ret->saddr, 0, sizeof(struct sockaddr_in));
	sin = (struct sockaddr_in *)ret->saddr;

	sin->sin_family = AF_INET;
	sin->sin_port = htons((uint16_t)port);
	memcpy(&sin->sin_addr.s_addr, &addr, 4);

	return ret;
}

/** Hostname resolution function */
Addr addr_create_from_hostname(const char *hostname, unsigned port)
{
#ifdef HAVE_GETADDRINFO 
	Addr ret;
	struct addrinfo hints, *res;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;

	if(getaddrinfo(hostname, 0, &hints, &res))
		return 0;

	ret = NEW(Addr);
	ret->saddr = (struct sockaddr *)xmalloc(sizeof(struct sockaddr_in));
	ret->slen = sizeof(struct sockaddr_in);

	memcpy(ret->saddr, res->ai_addr, sizeof(struct sockaddr_in));
	freeaddrinfo(res);

	((struct sockaddr_in *)ret->saddr)->sin_port = htons(port);

	return ret;
#else
        /* XXX We should confirm whether or not gethostbyname() uses 
           thread-specific data, and if not mutex lock the call (ouch) */
	Addr ret;
	struct sockaddr_in *sin;
	struct hostent *host;

	ret = NEW(Addr);
	ret->saddr = (struct sockaddr *)xmalloc(sizeof(struct sockaddr_in));
	ret->slen = sizeof(struct sockaddr_in);
	memset(ret->saddr, 0, sizeof(struct sockaddr_in));
	sin = (struct sockaddr_in *)ret->saddr;

	sin->sin_family = AF_INET;
	sin->sin_port = htons((uint16_t)port);

	if((int)(sin->sin_addr.s_addr = inet_addr(hostname)) == -1) {
		if(!(host = gethostbyname(hostname)) || host->h_length != 4) {
			xfree(ret);

			return 0;
		}

		memcpy(&sin->sin_addr, host->h_addr, 4);
#ifndef WIN32
		endhostent();
#endif
	}

	return ret;
#endif /* HAVE_GETADDRINFO */
}

void addr_destroy(Addr addr)
{
	xfree(addr->saddr);
	xfree(addr);
}

uint32_t addr_get_ipv4(Addr addr)
{
	if(((struct sockaddr_in *)addr->saddr)->sin_family != AF_INET)
		return 0;

	return (uint32_t)((struct sockaddr_in *)addr->saddr)->sin_addr.s_addr;
}

uint16_t addr_get_port(Addr addr)
{
	if(((struct sockaddr_in *)addr->saddr)->sin_family != AF_INET) 
		return 0;

	return ((struct sockaddr_in *)addr->saddr)->sin_port;
}

/** Connect function */
Sock sock_nbconnect_to_addr(Addr addr)
{
	Sock sock;

	if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return -1;

	if(set_blocking_mode(sock, SOCKET_NONBLOCK) < 0)
		goto err;

	if(connect(sock, addr->saddr, addr->slen) < 0) {
#ifdef WIN32
		if(WSAGetLastError() != WSAEWOULDBLOCK)
			goto err;
#else
		if(errno != EINPROGRESS)
			goto err;
#endif
	}

	return sock;
err:
	socket_close(sock);

	return -1;
}

int sock_nbconnect_complete(Sock sock)
{
	int errval;
	socklen_t errlen;

	errlen = sizeof(int);

	if(getsockopt(sock, SOL_SOCKET, SO_ERROR, (char *)&errval, &errlen) < 0)
		return -1;

	if(errval)
		return -1;

	return 0;
}

int set_blocking_mode(Sock sock, BlockingType mode)
{
	unsigned long flags;

#ifdef WIN32
	switch(mode) {
		case SOCKET_BLOCK:
			flags = 0;
			break;
		case SOCKET_NONBLOCK:
			flags = 1;
			break;
	}

	if(ioctlsocket(sock, FIONBIO, &flags) == SOCKET_ERROR)
		return -1;
#else
	if((flags = fcntl(sock, F_GETFL)) < 0)
		return -1;

	switch(mode) {
		case SOCKET_BLOCK:
			flags &= ~O_NONBLOCK;
			break;
		case SOCKET_NONBLOCK:
			flags |= O_NONBLOCK;
			break;
	}

	if(fcntl(sock, F_SETFL, flags) < 0)
		return -1;
#endif /* WIN32 */

	return 0;
}

Sock sock_connect_to_addr(Addr addr)
{
	Sock sock;
	fd_set sockset;
	struct timeval timeout;

	if((sock = sock_nbconnect_to_addr(addr)) < 0)
		return -1;

	FD_ZERO(&sockset);
	FD_SET(sock, &sockset);

	timeout.tv_sec = CONNECT_TIMEOUT;
	timeout.tv_usec = 0;

#ifdef WIN32
	if(select((int)sock + 1, 0, 0, &sockset, &timeout) < 1)
		goto err;
#else
	if(select(sock + 1, 0, &sockset, 0, &timeout) < 1)
		goto err;
#endif /* WIN32 */

	if(sock_nbconnect_complete(sock) < 0)
		goto err;

	return sock;

err:
	socket_close(sock);
	return -1;
}

Sock listener_ipv4_tcp_create(uint32_t source_addr, uint16_t port)
{
        Sock sock;
		int val = 1;
        struct sockaddr_in sin;

        memset(&sin, 0, sizeof(struct sockaddr_in));

        sin.sin_family = AF_INET;
        sin.sin_port = htons(port);
        sin.sin_addr.s_addr = source_addr;

        if((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP)) < 0) {
                perror("socket");
                return -1;
        }

        if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&val, sizeof(val)) <
 0) {
                perror("setsockopt");
                return -1;
        }

        if(bind(sock, (struct sockaddr *)&sin, sizeof(struct sockaddr_in)) < 0) {
                perror("bind");
                return -1;
        }

        if(listen(sock, MAX_BACKLOG) < 0) {
                perror("listen");
                return -1;
        }

        return sock;
}

Sock listener_ipv4_tcp_accept(Sock listensock, uint32_t *addr)
{
        Sock sock;
        struct sockaddr_in sin;
        unsigned int len = sizeof(struct sockaddr_in);

        if((sock = accept(listensock, (struct sockaddr *)&sin, &len)) < 0)
                return -1;

        if(addr)
                *addr = sin.sin_addr.s_addr;

        return sock;
}

Sock listener_ipv4_udp_create(uint32_t source_addr, uint16_t port)
{
        Sock sock;
        struct sockaddr_in sin;

        if((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
                perror("socket");
                return -1;
        }

        memset(&sin, 0, sizeof(struct sockaddr_in));
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = source_addr;
        sin.sin_port = htons(port);

        if(bind(sock, (struct sockaddr *)&sin, sizeof(struct sockaddr_in)) <
 0) {
                perror("bind");
                return -1;
        }

        return sock;
}

int listener_ipv4_udp_receive(Sock sock, void *buffer, unsigned length, uint32_t *addr)
{
        int ret;
        socklen_t len = sizeof(struct sockaddr_in);
        struct sockaddr_in sin;

        if((ret = recvfrom(sock, buffer, length, 0, (struct sockaddr *)&sin, &len)) < 0)
                return -1;

        if(addr)
                *addr = sin.sin_addr.s_addr;

        return ret;
}

/** Atomic read function with timeout */
ssize_t readsock_atomic(Sock sock, void *buf, size_t nbytes)
{
	fd_set sockset;
	struct timeval tv;

	tv.tv_sec = RESPONSE_TIMEOUT;
	tv.tv_usec = 0;

	FD_ZERO(&sockset);
	FD_SET(sock, &sockset);

	if(select((int)sock + 1, &sockset, 0, 0, &tv) < 1) 
		return -1;

	return recv(sock, buf, nbytes, 0);
}

/** Atomic write function with timeout */
ssize_t writesock_atomic(Sock sock, void *buf, size_t nbytes)
{
	fd_set sockset;
	struct timeval tv;

	tv.tv_sec = RESPONSE_TIMEOUT;
	tv.tv_usec = 0;

	FD_ZERO(&sockset);
	FD_SET(sock, &sockset);

	if(select((int)sock + 1, 0, &sockset, 0, &tv) < 1)
		return -1;

	return send(sock, buf, nbytes, 0);
}

int readsock(Sock sock, void *buf, size_t nbytes)
{
	ssize_t c;
	uint8_t *bptr = (uint8_t *)buf;

	while(nbytes > 0) {
		if((c = readsock_atomic(sock, bptr, nbytes)) < 1)
			return -1;

		nbytes -= c;
		bptr += c;
	}

	return 0;
}

int writesock(Sock sock, const void *buf, size_t nbytes)
{
	ssize_t c;
	uint8_t *bptr = (uint8_t *)buf;

	while(nbytes > 0) {
		if((c = writesock_atomic(sock, bptr, nbytes)) < 1)
			return -1;

		nbytes -= c;
		bptr += c;
	}

	return 0;
}

int readsock_int16(Sock sock, uint16_t *value)
{
	if(readsock(sock, value, 2) < 0)
		return -1;

	*value = ntohs(*value);

	return 0;
}

int readsock_int32(Sock sock, uint32_t *value)
{
	if(readsock(sock, value, 4) < 0)
		return -1;

	*value = ntohl(*value);

	return 0;
}

int writesock_int16(Sock sock, const uint16_t value)
{
	uint16_t v = htons(value);

	return writesock(sock, &v, 2);
}

int writesock_int32(Sock sock, const uint32_t value)
{
	uint32_t v = htonl(value);

	return writesock(sock, &v, 4);
}
