#include <CuTest.h>

#include "../Collection.h"

#include <assert.h>
#include <stdlib.h>

#define N_PTRS (sizeof(pointers) / sizeof *pointers)

static Collection c;
static int *pointers[17];
static int visited[N_PTRS];

/* helper functions */

static int collection_iterator1(CollectionKey key, void *value, void *data)
{
	int *ip = (int *)value;

	assert(key >= 0 && key < N_PTRS);
	assert(value != NULL);

	assert(*ip == key);
	assert(pointers[key] == ip);

	assert(data == (void *)0x57471C); /* correct data pointer? */

	visited[key] = key;

	return 1;
}

static int collection_iterator2(CollectionKey key, void *value, void *data)
{
	static int n_visits = 0;

	visited[key] = -1;

	return ++n_visits < N_PTRS / 2 ? 1 : 0;
}

static void collection_free_func(void *v)
{
	int *ip = (int *)v;

	assert(*ip < N_PTRS);
	assert(pointers[*ip] == ip);

	free(ip);
}

/* unit tests */

void test_collection_lifecycle1(CuTest *tc)
{
	c = collection_create();
	CuAssertPtrNotNull(tc, c);
	CuAssertIntEquals_Msg(tc, "Collection not empty upon create", 0, collection_size(c));
	collection_destroy(c, NULL);
}

void test_collection_lifecycle2(CuTest *tc)
{
	c = collection_create();
	CuAssert(tc, "can't insert into empty collection", collection_insert(c, 0xABCDEF, (void *)0xDEADBEEF) == 0);
	collection_destroy(c, NULL);
}

void test_collection_lifecycle3(CuTest *tc)
{
	int i, n_it, n_notit;

	c = collection_create();
	CuAssertPtrNotNull(tc, c);

	/* insert N_PTRS elements into the collection */
	for(i = 0; i < N_PTRS; i++) {
		pointers[i] = (int *)malloc(sizeof(int));
		*pointers[i] = i;
		CuAssert(tc, "error inserting into collection",
			collection_insert(c, i, pointers[i]) == 0);
	}
	CuAssertIntEquals(tc, N_PTRS, collection_size(c));

	for(i = 0; i < N_PTRS; i++)
		CuAssertPtrEquals(tc, pointers[i], collection_lookup(c, i));

	/* remove half the collection's elements */
	for(i = 0; i < N_PTRS / 2; i++)
		CuAssertPtrEquals(tc, pointers[i], collection_remove(c, i));

	for(i = 0; i < N_PTRS / 2; i++)
		CuAssertPtrEquals(tc, NULL, collection_lookup(c, i));
	for(; i < N_PTRS; i++)
		CuAssertPtrEquals(tc, pointers[i], collection_lookup(c, i));

	/* reinsert everything into the collection */
	for(i = 0; i < N_PTRS / 2; i++)
		CuAssert(tc, "error inserting into collection", collection_insert(c, i, pointers[i]) == 0);
	for(; i < N_PTRS; i++)
		CuAssert(tc, "no error inserting into collection", collection_insert(c, i, (void *)0xDEADBEEF) == -1);

	for(i = 0; i < N_PTRS; i++)
		CuAssertPtrEquals(tc, pointers[i], collection_lookup(c, i));

	/* iterate over the collection */
	for(i = 0; i < N_PTRS; i++)
		visited[i] = -1;
	collection_iterate(c, collection_iterator1, (void *)0x57471C);
	for(i = 0; i < N_PTRS; i++)
		CuAssertIntEquals_Msg(tc, "member not iterated over", i, visited[i]);

	for(i = 0; i < N_PTRS; i++)
		visited[i] = 0;
	collection_iterate(c, collection_iterator2, NULL);
	for(i = n_it = n_notit = 0; i < N_PTRS; i++) {
		if(visited[i])
			n_it++;
		else
			n_notit++;
	}
	CuAssertIntEquals_Msg(tc, "member(s) not iterated over", N_PTRS / 2, n_it);
	CuAssertIntEquals_Msg(tc, "member(s) iterated over", N_PTRS - N_PTRS / 2, n_notit);

	/* destroy the collection */
	CuAssertIntEquals(tc, N_PTRS, collection_size(c));

	collection_destroy(c, collection_free_func);
}
