#include <CuTest.h>

#include "../ComponentManager.h"

#include <stdio.h>

/* Component graph:

            c1   c6
           / \
          C2 c3
         / \ /
        C4 C5

 The components are listed as two DAGs, numbered in-order. If the components
 are created in the order given above, the componentmgr_{re}init_components()
 functions should visit the components in the order: c4, c5, c2, c3, c1, c6.
 The componentmgr_shutdown_components() function should visit the components
 in the opposite order, i.e. c6, c1, c3, c2, c5, c4.

 */

static Component c1, c2, c3, c4, c5, c6;
static int visit_count, prev_component;
static CuTest *tc;

static void c1_init(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 3, prev_component);
	prev_component = 1;
}

static void c1_reinit(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 3, prev_component);
	prev_component = 1;
}

static void c1_shutdown(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 6, prev_component);
	prev_component = 1;
}

static void c2_init(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 5, prev_component);
	prev_component = 2;
}

static void c2_reinit(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 5, prev_component);
	prev_component = 2;
}

static void c2_shutdown(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 3, prev_component);
	prev_component = 2;
}

static void c3_init(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 2, prev_component);
	prev_component = 3;
}

static void c3_reinit(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 2, prev_component);
	prev_component = 3;
}

static void c3_shutdown(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 1, prev_component);
	prev_component = 3;
}

static void c4_init(void)
{
	visit_count++;
	CuAssertIntEquals(tc, -1, prev_component);
	prev_component = 4;
}

static void c4_reinit(void)
{
	visit_count++;
	CuAssertIntEquals(tc, -1, prev_component);
	prev_component = 4;
}

static void c4_shutdown(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 5, prev_component);
	prev_component = 4;
}

static void c5_init(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 4, prev_component);
	prev_component = 5;
}

static void c5_reinit(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 4, prev_component);
	prev_component = 5;
}

static void c5_shutdown(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 2, prev_component);
	prev_component = 5;
}

static void c6_init(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 1, prev_component);
	prev_component = 6;
}

static void c6_reinit(void)
{
	visit_count++;
	CuAssertIntEquals(tc, 1, prev_component);
	prev_component = 6;
}

static void c6_shutdown(void)
{
	visit_count++;
	CuAssertIntEquals(tc, -1, prev_component);
	prev_component = 6;
}

void test_componentmgr(CuTest *cutest)
{
	tc = cutest;

	componentmgr_init();

	/* create components */
	c1 = component_create(c1_init, c1_reinit, c1_shutdown);
	c2 = component_create(c2_init, c2_reinit, c2_shutdown);
	c3 = component_create(c3_init, c3_reinit, c3_shutdown);
	c4 = component_create(c4_init, c4_reinit, c4_shutdown);
	c5 = component_create(c5_init, c5_reinit, c5_shutdown);
	c6 = component_create(c6_init, c6_reinit, c6_shutdown);

	/* register dependencies */
	component_register_dependency(c2, c5);
	component_register_dependency(c1, c3);
	component_register_dependency(c3, c5);
	component_register_dependency(c1, c2);
	component_register_dependency(c2, c4);

	visit_count = 0;
	prev_component = -1;
	componentmgr_init_components();
	CuAssertIntEquals(tc, 6, prev_component);
	CuAssertIntEquals(tc, 6, visit_count);

	visit_count = 0;
	prev_component = -1;
	componentmgr_reinit_components();
	CuAssertIntEquals(tc, 6, prev_component);
	CuAssertIntEquals(tc, 6, visit_count);

	visit_count = 0;
	prev_component = -1;
	componentmgr_shutdown_components();
	CuAssertIntEquals(tc, 4, prev_component);
	CuAssertIntEquals(tc, 6, visit_count);

	componentmgr_destroy();
}
