/**
  @file threadops.c 
  @brief Portable thread operations
*/
/*
   Copyright (C)2003-04 Anthony Arcieri
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#ifdef HAVE_PTHREADS
#include <pthread.h>
#endif /* HAVE_PTHREADS */

#include "threadops.h"
#include "xmalloc.h"

#if !defined(HAVE_PTHREADS) && !defined(WIN32)
#error Could not determine a suitable threads library for your platform
#endif

Thread thr_create(void *(*thread_func)(void *arg), void *arg)
{
#ifdef HAVE_PTHREADS
	Thread thread = (Thread)xmalloc(sizeof(pthread_t));
	
	if(pthread_create(thread, 0, thread_func, arg)) {
		xfree(thread);
		return 0;
	}

	return thread;
#endif /* HAVE_PTHREADS */

#ifdef WIN32
        return CreateThread(0, 0, (LPTHREAD_START_ROUTINE)thread_func, arg, 0, 0);
#endif /* WIN32 */
}

/** XXX Implement some means of reliably terminating threads */
void thr_destroy(Thread thread)
{
	/* XXX STUB! */
}

Mutex thr_mutex_create(void)
{
#ifdef HAVE_PTHREADS
        Mutex mutex = (Mutex)xmalloc(sizeof(pthread_mutex_t));
        pthread_mutex_init(mutex, 0);
        return mutex;
#endif /* HAVE_PTHREADS */

#ifdef WIN32
	Mutex mutex = (Mutex)xmalloc(sizeof(CRITICAL_SECTION));

	InitializeCriticalSection(mutex);
	
	return mutex;
#endif /* WIN32 */
}

void thr_mutex_destroy(Mutex mutex)
{
#ifdef HAVE_PTHREADS
	pthread_mutex_destroy(mutex);
	xfree(mutex);
#endif /* HAVE_PTHREADS */

#ifdef WIN32
	DeleteCriticalSection(mutex);
	xfree(mutex);
#endif
}

Cond thr_cond_create(void)
{
#ifdef HAVE_PTHREADS
	Cond cond = (Cond)xmalloc(sizeof(pthread_cond_t));
	pthread_cond_init(cond, 0);
	return cond;
#endif /* HAVE_PTHREADS */

#ifdef WIN32
	return CreateEvent(0, TRUE, FALSE, 0);
#endif /* WIN32 */
}

void thr_cond_destroy(Cond cond)
{
#ifdef HAVE_PTHREADS
	pthread_cond_destroy(cond);
	xfree(cond);
#endif /* HAVE_PTHREADS */

#ifdef WIN32
	CloseHandle(cond);
#endif /* WIN32 */
}

#ifdef WIN32
int thr_cond_wait(Cond cond, Mutex mutex)
{
	thr_mutex_unlock(mutex);

        if(WaitForSingleObject(cond, INFINITE) == WAIT_FAILED)
                return -1;

	thr_mutex_lock(mutex);

        return 0;
}
#endif /* WIN32 */
