/*
   @file util.c 
   @brief Miscellaneous utility functions
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

#include "util.h"
#include "xmalloc.h"

#ifdef WIN32
#define SLASH_TOKEN	"\\"
#else
#define SLASH_TOKEN	"/"
#endif

char *xinet_ntoa(uint32_t addr)
{
	char *ret = (char *)xmalloc(16);

	sprintf(ret, "%u.%u.%u.%u",  
			((uint8_t *)&addr)[0], 
			((uint8_t *)&addr)[1], 
			((uint8_t *)&addr)[2], 
			((uint8_t *)&addr)[3]);

	return ret;
}

/** Add a trailing platform-specific slash token to the end of a path */
void add_trailing_slash(char **path)
{
	size_t len;
	char *t;

	t = *path;

	/* XXX Should we really behave this way for the degenerate case? */
	if((len = strlen(t)) < 1) {
		xfree(t);
		*path = xstrdup(SLASH_TOKEN);
		return;
	}
	
	if(t[len - 1] == *SLASH_TOKEN)
		return;

	*path = (char *)xmalloc(len + 2);
	memcpy(*path, t, len);
	strcpy(*path + len, SLASH_TOKEN);
	xfree(t);
}

void get_systime(struct systime *st)
{
#ifdef WIN32
	SYSTEMTIME wst;

	GetLocalTime(&wst);

	st->year = wst.wYear;
	st->month = wst.wMonth;
	st->day = wst.wDay;
	st->hour = wst.wHour;
	st->minute = wst.wMinute;
	st->second = wst.wSecond;
#else
	time_t t;
	struct tm tm;
	
	time(&t);
	localtime_r(&t, &tm);

	st->year = tm.tm_year + 1900;
	st->month = tm.tm_mon + 1;
	st->day = tm.tm_mday;
	st->hour = tm.tm_hour;
	st->minute = tm.tm_min;
	st->second = tm.tm_sec;
#endif
}

/* Sleep the specified number of seconds */
void sleepsec(unsigned seconds)
{
#ifdef WIN32
	Sleep(seconds * 1000);
#else
	sleep(seconds);
#endif
}

/* Like strncpy, but always terminates */
void safecpy(char *dest, const char *src, unsigned length)
{
	strncpy(dest, src, length);
	dest[length - 1] = '\0';
}

/* Concatinate an arbitrary number of strings.  The last argument passed
   should always be zero to indicate the list's end.  Stores result in dest.
   If a null pointer is passed as dest allocate a sufficiently sized
   buffer to store the result and return it */
char *safecat(char *dest, unsigned length, ...)
{
	size_t l = 0;
    char *p, *str, *strp; 
    va_list ap; 

    va_start(ap, length); 
    while((p = (char *)va_arg(ap, char *))) 
            l += strlen(p); 
    va_end(ap); 

	strp = str = (char *)xmalloc(l + 1);

    va_start(ap, length);
    while((p = (char *)va_arg(ap, char *))) {
            l = strlen(p);
            memcpy(strp, p, l);
            strp += l;
    }
	va_end(ap);

	*strp = '\0';

	if(dest) {
		safecpy(dest, str, length);
		xfree(str);

		return dest;
	}
		
	return str;
}
