/**
  @file xmalloc.c 
  @brief Memory allocator/deallocator function wrappers
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>

#include "log.h"
#include "xmalloc.h"

/** Used by the portable vasprintf implementation */
#define INITIAL_BUFFER	128

/** On Windows, if a _vsnprintf() call fails we keep doubling the buffer size
   until we hit this sanity limit, to prevent an infinite allocation loop */
#define MAX_BUFFER 4096

void *xmalloc(size_t size)
{
	void *ret;

#ifdef WIN32
	if(!(ret = HeapAlloc(GetProcessHeap(), 0, size)))
#else
	if(!(ret = malloc(size)))
#endif
	{
		log_error("Couldn't allocate memory");
		fatal("Fatal: Memory management failure");
	}

	return ret;
}

void *xrealloc(void *ptr, size_t size)
{
	void *ret;

#ifdef WIN32
	if(!ptr)
		return xmalloc(size);

	if(!(ret = HeapReAlloc(GetProcessHeap(), 0, ptr, size)))
#else
	if(!(ret = realloc(ptr, size)))
#endif
	{
		log_error("Couldn't reallocate memory");
		fatal("Memory management failure", ret);
	}

	return ret;
}

char *xstrdup(const char *str)
{
	char *ret;
	
	ret = (char *)xmalloc(strlen(str) + 1);
	strcpy(ret, str);

	return ret;
}

char *xvasprintf(const char *format, va_list ap)
{
	char *ret;

#ifdef WIN32
        int size = INITIAL_BUFFER;

        ret = (char *)xmalloc(size);
        va_start(ap, format);

        while(_vsnprintf(ret, size, format, ap) < 0) {
                va_end(ap);
                xfree(ret);
 
		/* If we see this warning message we need to increase the 
		   size of MAX_BUFFER */
                if((size *= 2) > MAX_BUFFER)
			return xstrdup("[!WARNING!] Maximum string size of xvasprintf() exceeded");

                ret = (char *)xmalloc(size);
                va_start(ap, format);
        }

	va_end(ap);

	return ret;
#endif /* WIN32 */
	
#if HAVE_VASPRINTF
	if(vasprintf(&ret, format, ap) < 0) {
		log_error("vasprintf");
		fatal("Memory management failure");
	}

	return ret;
#endif /* HAVE_VASPRINTF */
	
#if !defined(WIN32) && !defined(HAVE_VASPRINTF)
	int l;

	ret = (char *)xmalloc(INITIAL_BUFFER);

	if((l = vsnprintf(ret, INITIAL_BUFFER, format, ap)) < INITIAL_BUFFER)  
		return ret;

	l++;
	xfree(ret);
	ret = (char *)xmalloc(l);

	vsnprintf(ret, l, format, ap);

	return ret;
#endif /* !defined(WIN32) && !defined(HAVE_VASPRINTF) */
}

char *xasprintf(const char *format, ...)
{
	char *ret;
	va_list ap;
		
	va_start(ap, format);
	ret = xvasprintf(format, ap);
	va_end(ap);

	return ret;
}
