/**
  @file HubConnection.c 
  @brief Code for connecting to and communicating with a hub
*/
/*
   Copyright (C)2004 Anthony Arcieri
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <string.h>

/* Common headers */
#include <Collection.h>
#include <Config.h>
#include <HMAC.h>
#include <Hasher.h>
#include <Lock.h>
#include <MsgQueue.h>
#include <Stack.h>
#include <Transaction.h>
#include <debug.h>
#include <log.h>
#include <msgval.h>
#include <socketops.h>
#include <threadops.h>
#include <txid.h>
#include <util.h>
#include <xmalloc.h>

/* Local headers */
#include "HubConnection.h"
#include "hubtxn.h"

/* Default port for PDTP servers */
#define DEFAULT_SOURCE_PORT	4044
#define DEFAULT_SLEEP_INTERVAL	10

/* Protocol magic numbers */
#define PDSP_MAGIC	0x50445350
#define AUTH_MAGIC	0x41555448

/** Authorization methods */
#define AUTH_HMAC_SHA1  0x0

static Sock hub_sock = -1;
static Lock hub_lock = 0;

static unsigned serial_counter = 1;
static Collection serial_collection = 0;
static Lock serial_lock = 0;
static Stack msgqueue_pool = 0;

static void hub_transaction_router_init()
{
	if(serial_lock)
		return;

	serial_lock = lock_create(LOCK_EXCL);
	serial_collection = collection_create();
	msgqueue_pool = stack_create();
}

static void hub_shutdown_notifier(void *queue)
{
	msgval val;

	val.ptr = 0;
	msgqueue_send((MsgQueue)queue, &val);
}

static void hub_transaction_router_destroy()
{
	MsgQueue queue;
	
	collection_destroy(serial_collection, hub_shutdown_notifier);

	/* XXX Possible race? */
	lock_destroy(serial_lock);
	
	while((queue = (MsgQueue)stack_pop(msgqueue_pool)))
		msgqueue_destroy(queue);

	stack_destroy(msgqueue_pool);
}

static MsgQueue hub_transaction_router_register(TransactionOut txn)
{
	uint32_t serial;
	MsgQueue queue;

	lock_writer_acquire(serial_lock);
	serial = serial_counter++;
	
	transaction_set_serial(txn, serial);
	
	if(!(queue = stack_pop(msgqueue_pool)))
		queue = msgqueue_create();

	collection_insert(serial_collection, serial, queue);
	lock_writer_release(serial_lock);

	return queue;
}

static TransactionIn hub_transaction_router_wait(MsgQueue queue)
{
	msgval val;

	msgqueue_recv(queue, &val);

	if(val.ptr) {
		lock_writer_acquire(serial_lock);
		stack_push(msgqueue_pool, queue);
		lock_writer_release(serial_lock);
	} else
		msgqueue_destroy(queue);

	return val.ptr;
}

static void hub_transaction_router_dispatch(TransactionIn txn)
{
	MsgQueue queue;
	msgval val;
	
	lock_writer_acquire(serial_lock);
	queue = collection_remove(serial_collection, transaction_serial(txn));
	lock_writer_release(serial_lock);
	
	if(!queue) {
		transaction_in_destroy(txn);
		return;
	}

	val.ptr = txn;
	msgqueue_send(queue, &val);
}

static int hub_auth_hmac_sha1(Sock sock, const char *key)
{
	uint8_t buf[20], hmac_digest[20], resp;

	if(readsock(sock, buf, 20) < 0)
		return -1;

	hmac_compute(PDTP_HASH_SHA1, key, (unsigned)strlen(key), buf, 20, hmac_digest);

	writesock(sock, hmac_digest, 20);

	if(readsock(sock, &resp, 1) < 0)
		return -1;

	/* Server must return ACK (0x1) or NACK(0x0) to conform to protocol */
	if(resp != 0 && resp != 1)
		return -1;

	return resp;
}

static int hub_auth(Sock sock)
{
	uint8_t resp;
	int ret;
	char *token;

	if((token = config_value("hub", "hmac_sha1_passphrase"))) {
		/* Select method 0x0: HMAC SHA1 authorization */
		writesock_int16(sock, AUTH_HMAC_SHA1);

		/* Check to determine if the hub allows this method */
		if(readsock(sock, &resp, 1) < 0) 
			return -1;

		/* If the hub doesn't allow this method, give up */
		if(resp != 0x1) 
			return -1;

		ret = hub_auth_hmac_sha1(sock, token);
		xfree(token);

		/* Check for error */
		if(ret < 0)
			return -1;

		/* Check for successful authentication */
		if(ret)
			return 0;
	}

	return -1;
}

static int hub_handshake(Sock sock)
{
	uint16_t v16;
	uint32_t v32;

	/* Send the magic number */
	writesock_int32(sock, PDSP_MAGIC);

	/* Read server's magic number */
	if(readsock_int32(sock, &v32) < 0)
		return -1;

	/* Check to see if server is requiring authorization */
	if(v32 == AUTH_MAGIC) {

		if(hub_auth(sock) < 0) 
			return -1;

		/* Read the new magic number */
		if(readsock_int32(sock, &v32) < 0)
			return -1;
	}

	if(v32 != PDSP_MAGIC)
		return -1;

	/* Read the minimum supported protocol version and check if it's 1 */
	if(readsock_int16(sock, &v16) < 0 || v16 != 1)
		return -1;

	/* Read the maximum supported protocol version */
	if(readsock_int16(sock, &v16) < 0)
		return -1;

	/* Select the protocol version, currently hardcoded to 1 */
	writesock_int16(sock, 1);

	return 0;
}

/** XXX This function and the entire module eventually needs to support 
   connecting to a list of PDTP servers.  Currently it only supports one */
static Sock hub_connect(const char *hostname, unsigned port)
{
	int sleep_interval;
	Sock sock;
	Addr addr;

	for(;;) {
		if(!(addr = addr_create_from_hostname(hostname, port)))
			fatal("!!! Error: Couldn't resolve '%s'", hostname);

		if((sock = sock_connect_to_addr(addr)) < 0)
			log_printf("!!! Error: Hub server %s:%d refused connection", hostname, port);

		addr_destroy(addr);

		if(sock < 0) 
			goto connect_error;

		if(!hub_handshake(sock))
			break;

		/* We've now encountered an error.  Sleep and reconnect */
		socket_close(sock);

		log_printf("!!! Error: %s:%d: Handshaking/authorization failed", hostname, port);
connect_error:
		if((sleep_interval = config_int_value("hub", "connect_interval")) < 0)
			sleep_interval = DEFAULT_SLEEP_INTERVAL;

		log_printf("*** Sleeping for %d seconds...", sleep_interval); 

		sleepsec(sleep_interval);
	}

	return sock;
}

/** XXX Temporary implementation until better event handling is implemented */
static void hub_transaction_wait()
{
	fd_set fds;
	TransactionIn txn;

	for(;;) {
		FD_ZERO(&fds);
		FD_SET(hub_sock, &fds);

		if(select((int)hub_sock + 1, &fds, 0, 0, 0) != 1)
			return;

		if(!(txn = transaction_read(hub_sock)))
			return;

		switch(transaction_type(txn)) {
			case TXN_COMMAND:
				if(hubtxn_exec(hub_sock, txn) < 0)
					return;

				break;
			case TXN_REPLY:
				hub_transaction_router_dispatch(txn);
				break;
		}
	}
}

static void *hubcxn_thread(void *arg)
{
	Sock sock;
	char *hostname, *t;
	unsigned port;

	for(;;) {
		if(!(hostname = config_value("hub", "address")))
			fatal("!!! Error: No hub::address defined in configuration");

		if(!(t = strchr(hostname, ':')))
			port = DEFAULT_SOURCE_PORT;
		else {
			*t++ = '\0';
			port = atoi(t);
		}

		sock = hub_connect(hostname, port);
		log_printf("*** Connected successfully to hub server at %s:%d", hostname, port);
		xfree(hostname);

		lock_writer_acquire(hub_lock);
		hub_sock = sock;
		lock_writer_release(hub_lock);

		hub_transaction_wait();

		log_printf("!!! Connection to hub lost!");

		lock_writer_acquire(hub_lock);
		socket_close(hub_sock);
		hub_sock = -1;
		lock_writer_release(hub_lock);
	}

	thr_exit(0);
}

void hubcxn_init()
{
	hub_transaction_router_init();
	
	/* XXX Add rehashing support */
	if(!hub_lock)
		hub_lock = lock_create(LOCK_SHARED);
	
	/* XXX Store the thread reference */
	if(!thr_create(hubcxn_thread, 0))
		fatal("!!! Could not start hub connection thread");
}

void hubcxn_destroy()
{
	hub_transaction_router_destroy();

	/* XXX STUB! */
	socket_close(hub_sock);
	hub_sock = -1;

	lock_destroy(hub_lock);
	hub_lock = 0;
}

/** Send a transaction which receives no reply */
int hubcxn_transact_async(TransactionOut txn)
{
	lock_reader_acquire(hub_lock);
	if(hub_sock == -1) {
		lock_reader_release(hub_lock);

		return -1;
	}

	transaction_write(hub_sock, txn);
	lock_reader_release(hub_lock);

	return 0;
}

/** Send a transaction and read the associated reply */
TransactionIn hubcxn_transact_sync(TransactionOut txn)
{
	MsgQueue queue;
	
	lock_reader_acquire(hub_lock);
	if(hub_sock == -1) {
		lock_reader_release(hub_lock);

		return 0;
	}

	queue = hub_transaction_router_register(txn);
	transaction_write(hub_sock, txn);
	lock_reader_release(hub_lock);

	return hub_transaction_router_wait(queue);
}
