/**
  @file PieceSet.c 
  @brief Bit vector piece set management.
*/
/*
   Copyright (C)2003-04 Matthew Bishop, Anthony Arcieri
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <assert.h>
#include <string.h>

/* Common headers */
#include <Config.h>
#include <RNG.h>
#include <log.h>
#include <xmalloc.h>

/* Local headers */
#include "PieceSet.h"

/** The maximum number of pieces must be <= 2 ** RANDOM_BYTES */
#define RANDOM_BYTES 2
#define MAX_RANDOM_BYTES 0xffff

/** number of bits in a member of the PieceSet::ps_bits array */
#define WORD_BITS (sizeof(int) * 8)
#define LAST_BIT (1 << (sizeof(int) - 1))

struct _PieceSet {
	int *ps_bits;
	int  ps_numwords;
	int  ps_numbits;
};

/** Lookup table that gives the number of 'on' bits in a byte. */
static char bitcount[] = {
	0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
	1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
	4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};

static unsigned ps_random_number(unsigned max)
{
        unsigned char buf[RANDOM_BYTES];
        unsigned r;

        rng_randbuf(buf, RANDOM_BYTES);

        r = buf[0];
        r |= (unsigned)buf[1] <<  8;

        return r / (MAX_RANDOM_BYTES / max);
}

PieceSet pieceset_create(int size)
{
	PieceSet set;

	set = (PieceSet)xmalloc(sizeof(struct _PieceSet));

	set->ps_numbits = size;

	set->ps_numwords = size / WORD_BITS + (size % WORD_BITS > 0 ? 1 : 0);

	set->ps_bits = (int *)xmalloc(set->ps_numwords);

	memset(set->ps_bits, 0, sizeof(int) * set->ps_numwords);

	return set;
}

static PieceSet pieceset_copysize(const PieceSet oldset)
{
	PieceSet set;

	set = (PieceSet)xmalloc(sizeof(struct _PieceSet));
	set->ps_bits = (int *)xmalloc(oldset->ps_numwords);
	set->ps_numbits = oldset->ps_numbits;
	set->ps_numwords = oldset->ps_numwords;

	return set;
}

void pieceset_destroy(PieceSet set)
{
	xfree(set->ps_bits);
	xfree(set);
}

int pieceset_get_piece(PieceSet set, int piece)
{
	int word, bit;

	assert(piece < set->ps_numbits);

	word = piece / WORD_BITS;
	bit = piece % WORD_BITS;

	return set->ps_bits[word] & (1 << bit);
}

void pieceset_add_piece(PieceSet set, int piece)
{
	int word, bit;

	assert(piece < set->ps_numbits);

	word = piece / WORD_BITS;
	bit = piece % WORD_BITS;

	set->ps_bits[word] |= 1 << bit;
}

void pieceset_add_range(PieceSet set, int start, int count)
{
	int word, end_word;
	int bit, end_bit, bits;

	assert(start + count <= set->ps_numbits);

	word = start / WORD_BITS;
	bit = 1 << (start % WORD_BITS);

	end_word = (start + count) / WORD_BITS;
	end_bit = 1 << ((start + count - 1) % WORD_BITS);

	while(word < end_word) {
		bits = bit;
		do {
			bit <<= 1;
			bits |= bit;
		} while(bit);
		set->ps_bits[word] |= bits;

		bit = 1;
		word++;
	}
	
/* lastword: */
	
	bits = bit;
	while(bit < end_bit) {
		bit <<= 1;
		bits |= bit;
	}
	set->ps_bits[word] |= bits;
}

void pieceset_remove_piece(PieceSet set, int piece)
{
	int word, bit;

	assert(piece < set->ps_numbits);

	word = piece / WORD_BITS;
	bit = piece % WORD_BITS;

	set->ps_bits[word] &= ~(1 << bit);
}

PieceSet pieceset_and(PieceSet set1, PieceSet set2)
{
	PieceSet result;
	int word;

	assert(set1->ps_numbits == set2->ps_numbits);

	result = pieceset_copysize(set1);

	for(word = 0; word < result->ps_numwords; word++)
		result->ps_bits[word] = set1->ps_bits[word] & set2->ps_bits[word];

	return result;
}

PieceSet pieceset_andv(PieceSet *sets, int numsets)
{
	PieceSet result;
	int i, word, numwords;
	int *resbits;

#ifndef NDEBUG
	{
		int i, n;

		for(i = 1, n = sets[0]->ps_numbits; i < numsets; i++)
			assert(sets[i]->ps_numbits == n);
	}
#endif /* NDEBUG */

	result = pieceset_copysize(sets[0]);
	memset(result->ps_bits, 0xff, result->ps_numwords * sizeof(int));
	resbits = result->ps_bits;
	numwords = sets[0]->ps_numwords;

	for(i = 0; i < numsets; i++) {
		int *bits = sets[i]->ps_bits;

		for(word = 0; word < numwords; word++)
			resbits[word] &= bits[word];
	}

	return result;
}

PieceSet pieceset_or(PieceSet set1, PieceSet set2)
{
	PieceSet result;
	int word;

	assert(set1->ps_numbits == set2->ps_numbits);

	result = pieceset_copysize(set1);

	for(word = 0; word < result->ps_numwords; word++)
		result->ps_bits[word] = set1->ps_bits[word] | set2->ps_bits[word];

	return result;
}

PieceSet pieceset_orv(PieceSet *sets, int numsets)
{
	PieceSet result;
	int i, word, numwords;
	int *resbits;

#ifndef NDEBUG
	{
		int i, n;

		for(i = 1, n = sets[0]->ps_numbits; i < numsets; i++)
			assert(sets[i]->ps_numbits == n);
	}
#endif /* NDEBUG */

	result = pieceset_copysize(sets[0]);
	memset(result->ps_bits, 0, result->ps_numwords * sizeof(int));
	resbits = result->ps_bits;
	numwords = sets[0]->ps_numwords;

	for(i = 0; i < numsets; i++) {
		int *bits = sets[i]->ps_bits;

		for(word = 0; word < numwords; word++)
			resbits[word] |= bits[word];
	}

	return result;
}

PieceSet pieceset_not(PieceSet set)
{
	PieceSet result;
	int word;

	result = pieceset_copysize(set);

	for(word = 0; word < result->ps_numwords; word++)
		result->ps_bits[word] = ~set->ps_bits[word];

	return result;
}

int pieceset_count(PieceSet set)
{
	int count = 0;
	int word, bit;

	for (word = 0; word < set->ps_numwords; word++) {
		for(bit = 0; bit < WORD_BITS; bit+=8)
			count += bitcount[(set->ps_bits[word] >> bit) & 0xff];
	}

	return count;
}

static int bit_n(int byte, int n)
{
	int i = 0, bit = 1;

	for(;;) {
		if(byte & bit)
			n--;

		if(n < 0)
			return i;

		bit <<= 1;
		i++;
	}
}

int pieceset_random_piece(PieceSet set)
{
	int word, bit, count, bits, n;

	n = ps_random_number(pieceset_count(set));

	count = 0;

	for (word = 0; word < set->ps_numwords; word++) {
		for(bit = 0; bit < WORD_BITS; bit+=8) {
			int i = (set->ps_bits[word] >> bit) & 0xff;
			bits =  bitcount[i];
			if(bits + count > n)
				return word * WORD_BITS + bit + bit_n(i, n - count);

			count += bits;
		}
	}

	fatal("!!! bug in random bit selector");

	/* This return is never reached but is here to suppress warnings */
	return -1;
}
