/**
  @file TransferManager.c 
  @brief State machine for all file transfers
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <assert.h>

/* Common headers */
#include <Collection.h>
#include <HashTable.h>
#include <Lock.h>
#include <debug.h>
#include <xmalloc.h>

/* Local headers */
#include "PieceSet.h"
#include "TransferManager.h"

typedef struct _FileData {
	char *path;
	uint32_t handle;
	uint32_t piece_count;
} *FileData;

Collection handle_table = 0;
Lock handle_table_lock = 0;

HashTable path_table = 0;
Lock path_table_lock = 0;

#define INITIALIZED	(handle_table && handle_table_lock)

void xfermgr_init(void)
{
	assert(!INITIALIZED);

	handle_table_lock = lock_create(LOCK_SHARED);
	handle_table = collection_create();

	path_table_lock = lock_create(LOCK_SHARED);
	path_table = hash_table_create();
}

void xfermgr_reinit(void)
{
	assert(INITIALIZED);
	
        lock_destroy(handle_table_lock);
        handle_table_lock = 0;

        collection_destroy(handle_table, 0);
        handle_table = 0;

        lock_destroy(path_table_lock);
        path_table_lock = 0;

        /* XXX Add destructor for path_table elements */
        hash_table_destroy(path_table, 0);
        path_table = 0;
}

void xfermgr_destroy(void)
{
	assert(INITIALIZED);

	lock_destroy(handle_table_lock);
	handle_table_lock = 0;
	
	/* XXX Add destructor for handle_table elements */
	collection_destroy(handle_table, 0);
	handle_table = 0;
}

void xfermgr_add_file(uint32_t handle, const char *path, uint32_t piece_count)
{
	FileData file;
	
	assert(INITIALIZED);
	
	file = NEW(FileData);
	file->handle = handle;
	file->path = xstrdup(path);
	file->piece_count = piece_count;

	lock_writer_acquire(handle_table_lock);
	collection_insert(handle_table, handle, file);
	lock_writer_release(handle_table_lock);

        lock_writer_acquire(path_table_lock);
        hash_table_insert(path_table, path, file);
        lock_writer_release(path_table_lock);

#ifndef NDEBUG
        debug("TransferManager: Adding file %s (%d) with %d pieces", path, handle, piece_count);
#endif
}

void xfermgr_remove_file(uint32_t handle)
{
	FileData file;

	assert(INITIALIZED);
	
#ifndef NDEBUG
	debug("TransferManager: Removing file %d", handle);
#endif

	lock_writer_acquire(handle_table_lock);
	file = collection_remove(handle_table, handle);
	lock_writer_release(handle_table_lock);

	if(!file) {
#ifndef NDEBUG
		debug("Received spurious file removal request for: %d", handle);
#endif
		return;
	}

        lock_writer_acquire(path_table_lock);
        hash_table_remove(path_table, file->path);
        lock_writer_release(path_table_lock);

	xfree(file->path);
	xfree(file);
}

int xfermgr_resolve_path(const char *path, uint32_t *handle)
{
        int ret = -1;
        FileData file;

        assert(INITIALIZED);

        lock_reader_acquire(path_table_lock);
        if(!(file = hash_table_lookup(path_table, path)))
                goto done;

        *handle = file->handle;
        ret = 0;
done:
        lock_reader_release(path_table_lock);

        return ret;
}

int xfermgr_get_piece_count(uint32_t handle, uint32_t *count)
{
	int ret = -1;
	FileData file;

	assert(INITIALIZED);

	lock_reader_acquire(handle_table_lock);
	if(!(file = collection_lookup(handle_table, handle)))
		goto done;

	*count = file->piece_count;
	ret = 0;
done:
	lock_reader_release(handle_table_lock);
	return ret;
}

void xfermgr_add_transfer(int uid, uint32_t file_handle, PieceSet remaining_pieces)
{
	assert(INITIALIZED);
}

void xfermgr_remove_transfer(int uid, uint32_t file_handle)
{
	assert(INITIALIZED);
}
