/**
  @file clienttxn.c 
  @brief Squall Server's handlers for client transactions
 */
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
 */

#include <machdep.h>

#include <string.h>

/* Common headers */
#include <Collection.h>
#include <Transaction.h>
#include <debug.h>
#include <transaction_replies.h>
#include <txid.h>
#include <util.h>
#include <xmalloc.h>

/* Local headers */
#include "HubConnection.h"
#include "PieceSet.h"
#include "TransferManager.h"
#include "clienttxn.h"

static int txn_download(int sock, TransactionIn txn);
static int txn_fwdtohub(int sock, TransactionIn txn);

struct th_list_element {
	uint16_t txn_id;
	int (*handler)(int sock, TransactionIn);
};

static struct th_list_element transaction_list[] = {
	{ TXN_DOWNLOAD, txn_download },
	{ TXN_STATDIR,	txn_fwdtohub },
	{ TXN_STATFILE,	txn_fwdtohub },
	{ TXN_DIRLIST,	txn_fwdtohub },
	{ TXN_FILEINFO,	txn_fwdtohub },
	{ TXN_DIRLIST,	txn_fwdtohub },
	{ 0, 0 }
};

static Collection transaction_table = 0;

void clienttxn_init()
{
	int i;

	transaction_table = collection_create();

	for(i = 0; transaction_list[i].txn_id; i++)
		collection_insert(transaction_table, (CollectionKey)transaction_list[i].txn_id, (void *)transaction_list[i].handler);
}

void clienttxn_destroy()
{
	collection_destroy(transaction_table, 0);
}

int clienttxn_exec(int sock, TransactionIn txn)
{ 
	int ret;
	int (*handler)(int, TransactionIn);

	if(!(handler = (int (*)(int, TransactionIn))collection_lookup(transaction_table, (CollectionKey)transaction_id(txn)))) {
		reply_error(sock, transaction_serial(txn), "Invalid transaction type (0x%x)", transaction_id(txn));
		return 0;
	}

	ret = handler(sock, txn);
	transaction_in_destroy(txn);

	return ret;
}

/* XXX POTENTIAL LOCKING BUGS - This implementation provides no way of
   checking to see if a given handle now represents a different file since
   the transfer began.  Higher level locking in the transfer manager is
   needed to remedy this problem */
static int txn_download(int sock, TransactionIn txn)
{ 
	int i, l;
	char *path;
	uint8_t *bptr;
	uint32_t handle, piece_count = 0, range_low, range_high, lastval = 0, v32;
	PieceSet set;

	if((i = transaction_object_index(txn, OBJ_PATH)) < 0) {
		reply_error(sock, transaction_serial(txn), "No path given");
		return 0;
	}

	path = transaction_object_string(txn, i);

	#ifndef NDEBUG
        debug("Client %d requested download of file %s", sock, path);
	#endif

	if(xfermgr_resolve_path(path, &handle) < 0 || xfermgr_get_piece_count(handle, &piece_count) < 0) {
		xfree(path);
		reply_error(sock, transaction_serial(txn), "File not found");
		return 0;
	}

	xfree(path);

	set = pieceset_create(piece_count);
	
	if((i = transaction_object_index(txn, OBJ_PIECELIST)) > 0) {
		l = transaction_object_length(txn, i);
		bptr = transaction_object_data(txn, i);

		/* Validate length of piece list */
		if(piece_count * 8 != l) {
			pieceset_destroy(set);
			reply_error(sock, transaction_serial(txn), "Invalid piece list length");
			return 0;
		}
		
		for(i = 0; i < piece_count; i++) {
			memcpy(&v32, bptr + i * 8, 4);
			range_low = ntohl(v32);

			memcpy(&v32, bptr + i * 8 + 4, 4);
			range_high = ntohl(v32);

			if(range_high < range_low || (i && lastval >= range_low)) {
				pieceset_destroy(set);
				reply_error(sock, transaction_serial(txn), "Malformed piece list");
				return 0;
			}
			
			pieceset_add_range(set, range_low, range_high);
			lastval = range_high;
		}
	}

	xfermgr_add_transfer(sock, 0, set);

	reply_success(sock, txn);

	/* XXX Dispatch event to transfer manager */

	return 0;
} 

static int txn_fwdtohub(int sock, TransactionIn txnin)
{
	uint32_t serial = transaction_serial(txnin);
	TransactionOut txnout = transaction_copy(txnin);

	txnin = hubcxn_transact_sync(txnout);
	transaction_out_destroy(txnout);

	if(!txnin) {
		reply_error(sock, serial, "No files currently available from this server (Disconnected from hub)");
		return 0;
	}

	txnout = transaction_copy(txnin);
	transaction_in_destroy(txnin);

	transaction_set_serial(txnout, serial);
	transaction_write(sock, txnout);
	transaction_out_destroy(txnout);

	return 0;
}
