/**
  @file main.c 
  @brief Squall PDTP server main program
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#ifdef HAVE_PTHREADS
#include <signal.h>
#include <pthread.h>
#endif

/* Common headers */
#include <ComponentManager.h>
#include <Config.h>
#include <Dispatcher.h>
#include <Listener.h>
#include <PoolCleaner.h>
#include <RNG.h>
#include <WorkerManager.h>
#include <log.h>
#include <util.h>
#include <xmalloc.h>

/* Local headers */
#include "ClientManager.h"
#include "HubConnection.h"
#include "TransferManager.h"
#include "clienttxn.h"
#include "config_keys.h"
#include "hubtxn.h"
#include "netevent.h"
#include "squall.h"

static void init_logging(void);
static void init_config(void);
#ifdef WIN32
static void init_sockets(void);
#else
static void init_signals(void);
#endif /* WIN32 */
	
/** Paths to search for configuration file */
#ifdef WIN32
static const char *config_file_path[] = {
	"squall.ini",
	"c:\\windows\\squall.ini",
	"c:\\winnt\\squall.ini",
	0
};

#else

static const char *config_file_path[] = { 
	"/var/squall/squall.conf",
	"/etc/squall.conf", 
	"/usr/local/etc/squall.conf",
	"squall.conf",
	0
};

#endif /* WIN32 */

static void component_init(void)
{
	Component Logging;
	Component PoolCleaner;
	Component TransferManager;
	Component ClientTransactionHandler;
	Component HubTransactionHandler;
	Component Dispatcher;
	Component ClientManager;
	Component Configuration;
	Component RNG;
	Component Listener;
	Component HubConnection;

#ifdef WIN32
        /* Win32-specific components */ 
	Component Sockets;
#else
	/* POSIX-specific components */
	Component Signals;
	Component WorkerManager;
#endif /* WIN32 */
	
	componentmgr_init();
	
	Configuration = component_create(init_config, init_config, config_destroy);

#ifdef WIN32
	Sockets = component_create(init_sockets, 0, 0);
#else
	Signals = component_create(init_signals, 0, 0);
#endif

	Logging = component_create(init_logging, 0, log_shutdown);

	PoolCleaner = component_create(poolcleaner_init, 0, poolcleaner_destroy);

	TransferManager = component_create(xfermgr_init, 0, xfermgr_destroy);
	
	ClientTransactionHandler = component_create(clienttxn_init, 0, clienttxn_destroy);

	HubTransactionHandler = component_create(hubtxn_init, 0, hubtxn_destroy);
	component_register_dependency(HubTransactionHandler, TransferManager);
	
#ifndef WIN32
	WorkerManager = component_create(workermgr_init, 0, workermgr_destroy);
	component_register_dependency(WorkerManager, Configuration);
#endif /* WIN32 */

	Dispatcher = component_create(dispatcher_init, 0, dispatcher_destroy);
	component_register_dependency(Dispatcher, ClientTransactionHandler);
#ifndef WIN32
	component_register_dependency(Dispatcher, WorkerManager);
#endif /* WIN32 */

	ClientManager = component_create(clientmgr_init, clientmgr_reinit, clientmgr_destroy);

	RNG = component_create(rng_init, 0, rng_destroy);
#ifndef WIN32
	component_register_dependency(RNG, Configuration);
#endif

	Listener = component_create(0, 0, listener_destroy);
	component_register_dependency(Listener, Configuration);

	HubConnection = component_create(hubcxn_init, 0, hubcxn_destroy);
	component_register_dependency(HubConnection, Configuration);
	component_register_dependency(HubConnection, HubTransactionHandler);

	componentmgr_init_components();

	/* XXX This is a temporary hack until a better interface to the listener
	   socket code.  Perhaps the component manager could be altered to
	   support this as well */
	listener_init(incoming_connection_handler, DEFAULT_PORT);
}

#ifdef WIN32
static void init_sockets()
{
        WSADATA data;
 
        /* Initialize Winsock */
        if(WSAStartup(MAKEWORD(2, 0), &data))
                fatal("!!! Couldn't initialize Winsock");
}
#else
static void init_signals(void)
{
        sigset_t sigset;

        /* Block signals we plan to handle or want to ignore */
        sigemptyset(&sigset);
        sigaddset(&sigset, SIGHUP);
        sigaddset(&sigset, SIGINT);
        sigaddset(&sigset, SIGTERM);
        sigaddset(&sigset, SIGPIPE);
	
        if(pthread_sigmask(SIG_BLOCK, &sigset, 0) < 0) {
		log_error("pthread_sigmask");
		fatal("!!! Couldn't set signal mask");
	}
}
#endif /* WIN32 */

static void init_logging(void)
{
	struct systime st;

	log_init();

	get_systime(&st);
	log_printf("*** %s %s started on %04d-%02d-%02d", PACKAGE, VERSION, st.year, st.month, st.day);
#ifndef WIN32
	log_printf("*** Built for %s on %s", PLATFORM, BUILDDATE);
#endif /* WIN32 */
}

/** Small wrapper for config_init, handles paths */
static void init_config(void)
{
	int i;

	for(i = 0; config_file_path[i]; i++) { 
		if(!config_init(config_file_path[i], config_keys)) {
			log_printf("*** Configuration successfully loaded from '%s'", config_file_path[i]);
			return;
		}
	}

	log_printf("!!! Warning: Couldn't open configuration file, using default settings");
}

static void event_loop(void)
{
#ifdef WIN32
	/* XXX Perhaps in the future we can find something better to do, but
	   for now this thread will be used to handle console control events
	   through a signal handler only, and will block here indefinately */

	WaitForSingleObject(GetCurrentThread(), INFINITE);
#else
	int signum;
	sigset_t sigset;

	sigemptyset(&sigset);
	sigaddset(&sigset, SIGHUP);
	sigaddset(&sigset, SIGINT);
	sigaddset(&sigset, SIGTERM);

	/* Handle SIGHUP */
	while(!sigwait(&sigset, &signum) && signum == SIGHUP) {
		log_printf("*** Server received SIGHUP, rehashing configuration"); 
		componentmgr_reinit_components();
	}

	log_printf("*** Server received signal %d, executing normal shutdown...", signum);

	/* We now unblock SIGINT and SIGTERM, just in case shutdown fails */
	sigemptyset(&sigset);
	sigaddset(&sigset, SIGINT);
	sigaddset(&sigset, SIGTERM);
	pthread_sigmask(SIG_UNBLOCK, &sigset, 0);
#endif

	componentmgr_shutdown_components();
	componentmgr_destroy();
}

int main(int argc, char **argv)
{
	/* Initialize all system components */
	component_init();

	/* Enter the main event loop */
        event_loop();

	/* On success exit with a successful return code */
        return 0;
}
