/**
  @file FileHasher.c
  @brief Generates list of MD5/SHA1 for pieces within a file
 */ 
/* 
   Copyright (C)2004 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <assert.h>

/* Common headers */
#include <Hasher.h>
#include <MsgQueue.h>
#include <debug.h>
#include <fileops.h>
#include <threadops.h>
#include <xmalloc.h>

/* Local headers */
#include "FileHasher.h"
#include "FileManager.h"

#define HASH_BLOCK_SIZE		16384
#define DEFAULT_PIECE_LENGTH	65536

int filehasher_generate_list(const char *path)
{
	FileHandle file;
	FileHashes hashes;
	Hasher hasher;
	HashAlgorithm cipher;
	unsigned piece_length = DEFAULT_PIECE_LENGTH;
	unsigned piece_count, hashbuffer_length, block_count, block_length;
        unsigned i, j, b;
		int n;
        uint8_t *hashbuffer, buffer[HASH_BLOCK_SIZE];
        uint64_t length, offset = 0;

#ifndef NDEBUG
        debug("Generating hashes for: %s", path);
#endif

	if(!(file = file_open(path))) {
#ifndef NDEBUG
                debug("!!! Error opening %s", path);
#endif
		
		filemgr_remove_file(path);
                return -1;
        }

	if(file_length(file, &length) < 0) {
		file_close(file);

#ifndef NDEBUG
		debug("!!! Couldn't determine length of %s", path);
#endif

		filemgr_remove_file(path);
		return -1;
	}

	cipher = PDTP_HASH_MD5;
	hasher = hasher_create(PDTP_HASH_MD5);
	
        piece_count = (unsigned)(length / piece_length);

        if(length % piece_length)
                piece_count++;

        hashbuffer_length = hasher_digest_size(hasher) * piece_count;
        hashbuffer = (uint8_t *)xmalloc(hashbuffer_length);

        for(i = 0; i < piece_count; i++) {
                hasher_zero(hasher);
                block_count = piece_length / HASH_BLOCK_SIZE;

                if(piece_length % HASH_BLOCK_SIZE)
                        block_count++;

                block_length = piece_length;

                for(j = 0; j < block_count; j++) {
                        b = block_length > HASH_BLOCK_SIZE ? HASH_BLOCK_SIZE : block_length;
                        b = (unsigned)(b > length ? length : b);

                        if((n = file_read(file, buffer, b, offset)) < 0) {
                                file_close(file);
                                hasher_destroy(hasher);
                                xfree(hashbuffer);

                                return -1;
                        }

                        hasher_add_data(hasher, buffer, n);
			
                        length -= n;
                        block_length -= n;
			offset += n;
                }

                hasher_get_digest(hasher, hashbuffer + (hasher_digest_size(hasher) * i));
        }

        file_close(file);
        hasher_destroy(hasher);

	hashes = filehashes_create(cipher, piece_length, hashbuffer, hashbuffer_length);
	xfree(hashbuffer);

	filemgr_set_hashes(path, hashes);
	filehashes_destroy(hashes);

        return 0;
}

void filehasher_init(void)
{
#ifndef NDEBUG
	debug("File hasher starting");
#endif
}

void filehasher_destroy(void)
{
#ifndef NDEBUG
	debug("File hasher stopping");
#endif
}


void filehasher_enqueue(const char *path)
{
	filehasher_generate_list(path);
}

void filehasher_dequeue(const char *path)
{
}
