/**
  @file FileManager.c 
  @brief Manages files being served
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <assert.h>

/* Common headers */
#include <Collection.h>
#include <Indexer.h>
#include <HashTable.h>
#include <Hasher.h>
#include <Lock.h>
#include <Transaction.h>
#include <debug.h>
#include <log.h>
#include <txid.h>
#include <xmalloc.h>

/* Local headers */
#include "FileEntry.h"
#include "FileHasher.h"
#include "FileHashes.h"
#include "FileManager.h"
#include "ServerManager.h"
#include "VFSMonitor.h"

static Collection handle_table = 0;
static Lock handle_table_lock = 0;

static HashTable path_table = 0;
static Lock path_table_lock = 0;

static Indexer handle_pool = 0;
static Lock handle_pool_lock = 0;

#define INITIALIZED	(handle_table_lock && handle_table && path_table_lock && path_table && handle_pool_lock && handle_pool)

#define MIN_HANDLE	1
#define MAX_HANDLE	0xFFFFFFFF

/* Static functions to inform servers about newly available/unavailable files */
static void broadcast_addhandle(uint32_t handle, const char *path, uint32_t piece_count)
{
        TransactionOut txn;
	char *pdtp_path;
        
        if(!(pdtp_path = vfsmonitor_hton_path(path))) {
#ifndef NDEBUG
                debug("Warning: got NIL pdtp_path, skipping...");
#endif
                return;
        }

        txn = transaction_create(TXN_COMMAND, TXN_ADDHANDLE, 3);
        transaction_add_int32(txn, OBJ_HANDLE, handle);
	transaction_add_string(txn, OBJ_PATH, pdtp_path);
        transaction_add_int32(txn, OBJ_PIECEID, piece_count);

        servermgr_broadcast(txn);
        transaction_out_destroy(txn);

	xfree(pdtp_path);
}

static void broadcast_remhandle(uint32_t handle)
{
        TransactionOut txn;

        txn = transaction_create(TXN_COMMAND, TXN_REMHANDLE, 1);
        transaction_add_int32(txn, OBJ_HANDLE, handle);

        servermgr_broadcast(txn);
        transaction_out_destroy(txn);
}

/* FileManager routines */
void filemgr_init(void)
{
	assert(!INITIALIZED);
	
	handle_table_lock = lock_create(LOCK_SHARED);
	handle_table = collection_create();
			
	path_table_lock = lock_create(LOCK_SHARED);
	path_table = hash_table_create();

	handle_pool_lock = lock_create(LOCK_EXCL);
	handle_pool = indexer_create(MIN_HANDLE, MAX_HANDLE);
}

void filemgr_reinit(void)
{
	assert(INITIALIZED);
}

void filemgr_destroy(void)
{
	assert(INITIALIZED);
	
	lock_destroy(handle_table_lock);
	handle_table_lock = 0;

	collection_destroy(handle_table, 0);
	handle_table = 0;

	lock_destroy(path_table_lock);
	path_table_lock = 0;
	
	/* XXX Add destructor for path_table elements */
	hash_table_destroy(path_table, 0);
	path_table = 0;

	lock_destroy(handle_pool_lock);
	handle_pool_lock = 0;

	indexer_destroy(handle_pool);
	handle_pool = 0;
}

void filemgr_add_file(const char *path)
{
	FileEntry file;
	uint32_t handle;

	assert(INITIALIZED);

	lock_writer_acquire(handle_pool_lock);
	if(indexer_get(handle_pool, &handle) < 0) {
		log_printf("!!! Warning: Couldn't register %s due to insufficient file handles", path);
		lock_writer_release(handle_pool_lock);

		return;
	}
	lock_writer_release(handle_pool_lock);
	
	if(!(file = fileentry_create(path, handle))) {
		log_error(path);
		return;
	}

	lock_writer_acquire(path_table_lock);
	hash_table_insert(path_table, path, file);
	lock_writer_release(path_table_lock);
	
	lock_writer_acquire(handle_table_lock);
	collection_insert(handle_table, handle, file);
	lock_writer_release(handle_table_lock);

#ifndef NDEBUG
	debug("FileManager: Added %s as handle %d", path, handle);
#endif

	filehasher_enqueue(path);
}

void filemgr_remove_file(const char *path)
{
	FileEntry file;
	uint32_t handle;

	assert(INITIALIZED);
	
	lock_writer_acquire(path_table_lock);
	file = hash_table_remove(path_table, path);
	lock_writer_release(path_table_lock);

	if(!file) {
#ifdef DEBUG
		debug("!!! Warning: Spurious VFS event for %s", path);
#endif
		return;
	}

	handle = fileentry_get_handle(file);
	broadcast_remhandle(handle);

        lock_writer_acquire(handle_table_lock);
        collection_remove(handle_table, handle);
        lock_writer_release(handle_table_lock);

	fileentry_destroy(file);
	
#ifndef NDEBUG
	debug("FileManager: Removed %s", path);
#endif

	filehasher_dequeue(path);

	lock_writer_acquire(handle_pool_lock);
        indexer_return(handle_pool, handle);
	lock_writer_release(handle_pool_lock);
}

int filemgr_set_hashes(const char *path, FileHashes hashes)
{
	FileEntry file;
	uint32_t handle;
	uint32_t piece_count;
	
	assert(INITIALIZED);

	lock_writer_acquire(path_table_lock);

	if(!(file = hash_table_lookup(path_table, path))) {
		lock_writer_release(path_table_lock);
		return -1;
	}
	
	handle = fileentry_get_handle(file);
	piece_count = filehashes_get_piece_count(hashes);

	fileentry_set_hashes(file, hashes);
	lock_writer_release(path_table_lock);

	broadcast_addhandle(handle, path, piece_count);

	return 0;
}

FileEntry filemgr_get_entry_by_path(const char *path, LockType type)
{
	FileEntry file, copy;
	
	assert(INITIALIZED);

	lock_reader_acquire(path_table_lock);

	if(!(file = hash_table_lookup(path_table, path))) {
		lock_reader_release(path_table_lock);
		return 0;
	}

	copy = fileentry_copy(file);
	lock_reader_release(path_table_lock);

	return copy;
}

FileEntry filemgr_get_entry_by_handle(uint32_t handle, LockType type)
{
	assert(INITIALIZED);

	return 0;
}

void filemgr_return_entry(FileEntry entry)
{
	fileentry_destroy(entry);
}

static int blast_handles_iterator(void *entryptr, void *sockptr)
{
	int sock = *(int *)sockptr;
	char *pdtp_path;
	FileEntry file = (FileEntry)entryptr;
	FileHashes hashes;
	TransactionOut txn;
	
	if(!(hashes = fileentry_get_hashes(file)))
		return -1;
	
        txn = transaction_create(TXN_COMMAND, TXN_ADDHANDLE, 3);
        transaction_add_int32(txn, OBJ_HANDLE, fileentry_get_handle(file));

	pdtp_path = vfsmonitor_hton_path(fileentry_get_path(file));
	assert(pdtp_path);
	transaction_add_string(txn, OBJ_PATH, pdtp_path);
	xfree(pdtp_path);

        transaction_add_int32(txn, OBJ_PIECEID, filehashes_get_piece_count(hashes));

	transaction_write(sock, txn);
        transaction_out_destroy(txn);
	filehashes_destroy(hashes);

	return 1;
}

void filemgr_blast_handles(int sock)
{
#ifndef NDEBUG
	debug("Blasting handle list to newly connected server");
#endif

	lock_reader_acquire(path_table_lock);
	hash_table_iterate(path_table, blast_handles_iterator, &sock);
	lock_reader_release(path_table_lock);
}
