/*
 @file ServerManager.c 
 @brief Stores information about all connected servers and proxies 
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <assert.h>

/* Common headers */
#include <Collection.h>
#include <Lock.h>
#include <debug.h>
#include <log.h>
#include <util.h>
#include <xmalloc.h>

/* Local headers */
#include "FileManager.h"
#include "Server.h"
#include "ServerManager.h"

static Collection server_table = 0;
static Lock server_table_lock = 0;

#define INITIALIZED	(server_table && server_table_lock)

void servermgr_init()
{
	assert(!INITIALIZED);

	server_table_lock = lock_create(LOCK_SHARED);
	server_table = collection_create();
}

void servermgr_reinit()
{
	assert(INITIALIZED);
}

void servermgr_destroy()
{
	assert(INITIALIZED);

	lock_destroy(server_table_lock);
	server_table_lock = 0;

	collection_destroy(server_table, (void (*)(void *))server_destroy);
	server_table = 0;
}

int servermgr_add_ipv4_connection(int sock, uint32_t addr)
{
	Server server;
	char *t;
	
	assert(INITIALIZED);
	
	log_printf("*** %s has connected as a PDTP server", t = xinet_ntoa(addr));
	xfree(t);

	server = server_create(sock, addr);

	lock_writer_acquire(server_table_lock);
	collection_insert(server_table, sock, server);
	lock_writer_release(server_table_lock);

	filemgr_blast_handles(sock);

	return 0;
}

void servermgr_remove_connection(int sock)
{
	Server server;
	char *t;
	
	assert(INITIALIZED);

	lock_writer_acquire(server_table_lock);
	server = (Server)collection_remove(server_table, sock);
	lock_writer_release(server_table_lock);

	/* Check to see internal structures are consistent */
	assert(server != 0);

	log_printf("--- Server %s disconnected", t = xinet_ntoa(server_addr(server)));
	xfree(t);

	server_destroy(server);
}

static int broadcast_iterator(CollectionKey key, void *server, void *txn)
{
	transaction_write(server_sock((Server)server), (TransactionOut)txn); 

	return 1;
}

void servermgr_broadcast(TransactionOut txn)
{
	assert(INITIALIZED);

	lock_reader_acquire(server_table_lock);
	collection_iterate(server_table, broadcast_iterator, txn);
	lock_reader_release(server_table_lock);
}
