/**
  @file VFSMonitor.c 
  @brief Abstract VFS monitoring interface

  A QUICK EXPLANATION OF THESE FUNCTIONS:

  These functions attempt to implement abstract filesystem monitoring
  while centralizing as much code as possible, only utilizing platform
  specific functionality where necessary.  Files are tracked not by
  location, but as sets of specific data.  If a file is altered, it should
  be considered "lost" before being "found" again (i.e. one set of specific
  data has disappeared, but a new set has been found at the same location).

  XXX Internal directory heirarchies are still not being represented correctly
  and references to freed pointers are still being generated.  More extensive
  debugging is necessary to remedy these problems.

  XXX Error handling within the VFS monitor thread itself is virtually
  non-existant at this point.  A defined behavior for handling errors within
  this thread needs to be implemented.
 */
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

#include <assert.h>
#include <string.h>

#ifndef WIN32
#include <limits.h>
#endif /* WIN32 */

/* Common headers */
#include <Config.h>
#include <DirEntry.h>
#include <HashTable.h>
#include <Lock.h>
#include <SysDir.h>
#include <debug.h>
#include <global.h>
#include <log.h>
#include <threadops.h>
#include <util.h>
#include <xmalloc.h>

/* Local headers */
#include "DirList.h"
#include "FileManager.h"
#include "VFSMonitor.h"

typedef struct _FileNode {
	char *path;
	char *name;
	uint64_t length;
	uint32_t ts_year;
	uint32_t ts_ysec;
} *FileNode;

typedef struct _FileListNode {
	FileNode node;
	struct _FileListNode *next;
} *FileListNode;

typedef struct _DirNode {
#ifndef WIN32
	int fd;
#endif
	char *path;
	char *name;
	uint32_t ts_year;
	uint32_t ts_ysec;

	unsigned subdir_count, file_count;
	struct _DirNode **subdirs;
	struct _FileNode **files;
} *DirNode;

typedef struct _DirListNode {
	DirNode node;
	struct _DirListNode *next;
} *DirListNode;

static FileNode filenode_create(const char *path);
static void filenode_destroy(FileNode node);

static DirNode dirnode_create(const char *path);
static void dirnode_destroy(DirNode node);
static void dirnode_event(DirNode node, int recursive);

static void vfs_scan_complete(const char *path);

/* Global state variables */
static Lock monitor_lock = 0;
static DirNode monitor_root = 0;
static HashTable dirnode_table = 0;

static Lock base_path_lock = 0;
static char *base_path = 0;

/* Win32 ReadDirectoryChangesW() VFS monitor implementation */
#ifdef WIN32
#define HAVE_VFS_MONITOR 	1
#define NOTIFY_BUFFER_SIZE      1024 /* Size of directory event buffer */

#define dirnode_monitor(node) 	0	
#define dirnode_unmonitor(node)	0
#define vfsmonitor_local_init()	0
#define vfsmonitor_local_shutdown()

static void vfsmonitor_event_loop(const char *ppath)
{
	size_t l, ppl;
	char *buffer[NOTIFY_BUFFER_SIZE] = { 0 }, path[PDTP_PATH_MAX + 1], *pp, *t;
	HANDLE dir;
	DWORD bytesread;
	PFILE_NOTIFY_INFORMATION info = (PFILE_NOTIFY_INFORMATION)buffer;

	DirNode node;

	safecpy(path, ppath, PDTP_PATH_MAX);
	l = strlen(path);

	if(!l || path[l - 1] != '\\') {
		strcpy(path + l, "\\");
		l++;
	}

#ifndef NDEBUG
	debug("Creating filesystem monitor for: '%s'", path);
#endif

	if((dir = CreateFile(path, FILE_LIST_DIRECTORY, FILE_SHARE_READ | FILE_SHARE_DELETE, 0, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0)) == INVALID_HANDLE_VALUE) {
#ifndef NDEBUG
		debug("Couldn't open %s with CreateFile()", path);
#endif
		/* XXX Better error handling is in order here */
		thr_exit(0);
	}

	pp = path + l;

	if((ppl = PDTP_PATH_MAX - l) < 0) {
#ifndef NDEBUG
		debug("Path too long: %s", path);
#endif
		/* XXX Better error handling is in order here */
		goto done;
	}

	lock_writer_acquire(monitor_lock);

	if(!(monitor_root = dirnode_create(path))) {
#ifndef NDEBUG
		debug("Processing of %s failed", path);
#endif
		/* XXX Better error handling is in order here */
		goto done;
	}

	lock_writer_release(monitor_lock);

	vfs_scan_complete(path);

	while(ReadDirectoryChangesW(dir, buffer, NOTIFY_BUFFER_SIZE, TRUE, FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_DIR_NAME | FILE_NOTIFY_CHANGE_ATTRIBUTES | FILE_NOTIFY_CHANGE_SIZE | FILE_NOTIFY_CHANGE_LAST_WRITE | FILE_NOTIFY_CHANGE_CREATION, &bytesread, 0, 0)) {
		if(!bytesread) {
			lock_writer_acquire(monitor_lock);
			dirnode_event(monitor_root, 1);
			lock_writer_release(monitor_lock);

			continue;
		}

		if(!WideCharToMultiByte(CP_ACP, 0, info->FileName, info->FileNameLength, pp, (int)ppl, 0, 0))
			continue;

		/* Pay attention only to parent directories */
		if((t = strrchr(path, '\\')))
			t[1] = '\0';

		lock_writer_acquire(monitor_lock);
		if((node = hash_table_lookup(dirnode_table, path)))
			dirnode_event(node, 0);
		lock_writer_release(monitor_lock);

		/* XXX For some reason, FileNameLength does not seem to get set
		   properly unless we zero between successive calls.  Someone
		   with Win32 experience should probably audit this code
		   and ensure that it's written correctly */
		ZeroMemory(buffer, NOTIFY_BUFFER_SIZE);
	} 

done: 
	CloseHandle(dir);
}

#endif /* WIN32 */

/* Linux dnotify/F_NOTIFY VFS monitor implementation */
#ifdef HAVE_FNOTIFY
#define HAVE_VFS_MONITOR 1

#include <signal.h>
#include <Collection.h>

static Collection monitor_collection = 0;

static int dirnode_monitor(DirNode node)
{
	if(fcntl(node->fd, F_SETSIG, SIGRTMIN) < 0)
		return -1;

	if(fcntl(node->fd, F_NOTIFY, DN_MODIFY | DN_CREATE | DN_DELETE | DN_RENAME | DN_ATTRIB | DN_MULTISHOT) < 0)
		return -1;

	collection_insert(monitor_collection, node->fd, node);

	return 0;
}

static int dirnode_unmonitor(DirNode node)
{
	collection_remove(monitor_collection, node->fd);

	return 0;
}

static int vfsmonitor_local_init(void)
{
	struct sigaction sigact;
	sigset_t sigset;

	/* Handle signal initialization the first time */
	sigemptyset(&sigset);
	sigaddset(&sigset, SIGRTMIN);

	if(pthread_sigmask(SIG_UNBLOCK, &sigset, 0) < 0)
		fatal("Couldn't set thread's signal mask");

	sigact.sa_sigaction = 0;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = SA_SIGINFO | SA_RESTART | SA_NODEFER;

	if(sigaction(SIGRTMIN, &sigact, 0) < 0)
		fatal("Couldn't initialize VFS monitor signal handling");

	if(pthread_sigmask(SIG_BLOCK, &sigset, 0) < 0)
		fatal("Couldn't set thread's signal mask");

	monitor_collection = collection_create();

	return 0;
}

static void vfsmonitor_local_shutdown(void)
{
	collection_destroy(monitor_collection, 0);
}

static void vfsmonitor_event_loop(const char *path)
{
	sigset_t sigset;
	siginfo_t siginfo;
	DirNode node;

	sigemptyset(&sigset);
	sigaddset(&sigset, SIGRTMIN);

#ifndef NDEBUG
	debug("+++ F_NOTIFY VFS monitor event loop started");
#endif

	lock_writer_acquire(monitor_lock);
	if(!(monitor_root = dirnode_create(path)))
		fatal("Couldn't initialize filesystem monitor");
	lock_writer_release(monitor_lock);

	vfs_scan_complete(path);

	while(sigwaitinfo(&sigset, &siginfo) > -1) {
		lock_writer_acquire(monitor_lock);

		if((node = (DirNode)collection_lookup(monitor_collection, siginfo.si_fd)))
			dirnode_event(node, 0);

		lock_writer_release(monitor_lock);
	}

#ifndef NDEBUG
	debug("--- F_NOTIFY VFS monitor event loop ended");
#endif
}

#endif /* HAVE_FNOTIFY */

/* xBSD kqueue implementation */
#ifdef HAVE_KQUEUE
#define HAVE_VFS_MONITOR 1

#include <sys/event.h>

static int monitorfd = -1;

static int dirnode_monitor(DirNode node)
{
	struct kevent ev;

	EV_SET(&ev, node->fd, EVFILT_VNODE, EV_ADD | EV_ENABLE | EV_CLEAR, NOTE_DELETE | NOTE_WRITE | NOTE_ATTRIB | NOTE_RENAME | NOTE_REVOKE, 0, node);

	if(kevent(monitorfd, &ev, 1, 0, 0, 0) < 0) {
		log_error("kevent");
		return -1;
	}

	return 0;
}

static int dirnode_unmonitor(DirNode node)
{
	struct kevent ev;

	EV_SET(&ev, node->fd, EVFILT_VNODE, EV_DELETE, 0, 0, 0);

	if(kevent(monitorfd, &ev, 1, 0, 0, 0) < 0) {
		log_error("kevent");
		return -1;
	}

	return 0;
}

static int vfsmonitor_local_init(void)
{
	if((monitorfd = kqueue()) < 0) 
		return -1;

	return 0;
}

static void vfsmonitor_local_shutdown(void)
{
	close(monitorfd);
}

static void vfsmonitor_event_loop(const char *path)
{
	struct kevent ev;

	lock_writer_acquire(monitor_lock);
	if(!(monitor_root = dirnode_create(path)))
		fatal("Couldn't initialize filesystem monitor");
	lock_writer_release(monitor_lock);

	vfs_scan_complete(path);

#ifndef NDEBUG
	debug("+++ kqueue() VFS monitor event loop started");
#endif

	while(kevent(monitorfd, 0, 0, &ev, 1, 0) == 1) {
		lock_writer_acquire(monitor_lock);
		dirnode_event((DirNode)ev.udata, 0);
		lock_writer_release(monitor_lock);
	}

#ifndef NDEBUG
	debug("--- kqueue() VFS monitor event loop ended");
#endif
}

#endif /* HAVE_KQUEUE */

/* Null macros for hosts without VFS monitoring */
#ifndef HAVE_VFS_MONITOR

#define dirnode_monitor(node) 		0
#define dirnode_unmonitor(node) 	0

#define vfsmonitor_local_init() 	0
#define vfsmonitor_local_shutdown()

#define vfsmonitor_event_loop(path)

#endif

static FileNode filenode_create(const char *path)
{
	FileNode node;
	struct DirEntry entry;

	if(direntry_get(path, &entry) < 0)
		return 0;

	node = NEW(FileNode);

	node->path = xstrdup(path);
	node->name = xstrdup(entry.name);
	node->length = entry.length;
	node->ts_year = entry.ts_year;
	node->ts_ysec = entry.ts_ysec;

	filemgr_add_file(node->path);

	return node;
}

static void filenode_destroy(FileNode node)
{
	filemgr_remove_file(node->path);

	xfree(node->path);
	xfree(node->name);
	xfree(node);
}

static DirNode dirnode_create(const char *path)
{
	size_t i,l;
	SysDir dir;
	struct DirEntry entry;
#ifdef WIN32
	char subpath[PDTP_PATH_MAX];
	const unsigned SUBPATHLEN = PDTP_PATH_MAX;
#else
	unsigned dotfiles = 1;
	char subpath[PATH_MAX];
	const unsigned SUBPATHLEN = PATH_MAX;
#endif /* WIN32 */

	DirListNode dirlist = 0, dltmp;
	FileListNode filelist = 0, fltmp;
	DirNode dn, node = NEW(DirNode);
	FileNode fn;

#ifndef WIN32
	if(config_truth_value("files", "ignore_dotfiles"))
		dotfiles = 0;
#endif /* WIN32 */

	node->subdir_count = 0;
	node->file_count = 0;

	if(direntry_get(path, &entry) < 0)
		goto err0;

#if defined(HAVE_KQUEUE) || defined(HAVE_FNOTIFY)
	if((node->fd = open(path, O_RDONLY)) < 0)
		goto err0;
#endif /* defined(HAVE_KQUEUE) || defined(HAVE_FNOTIFY) */

	if(!(dir = sysdir_open(path)))
		goto err1;

	if(dirnode_monitor(node) < 0)
		goto err2;

	node->path = xstrdup(path);
	add_trailing_slash(&node->path);

	hash_table_insert(dirnode_table, node->path, node);

#ifndef NDEBUG
	debug("Tracking directory: %s", node->path);
#endif
	hash_table_insert(dirnode_table, node->path, node);

	node->name = xstrdup(entry.name);
	node->ts_year = entry.ts_year;
	node->ts_ysec = entry.ts_ysec;

	while(!sysdir_get_entry(dir, &entry)) {
		if(*entry.name == '.') {
#ifndef WIN32
			if(!dotfiles)
				continue;
#endif

			l = strlen(entry.name);

			/* Ignore '.' and '..' */
			if(l == 1)
				continue;

			if(l == 2 && entry.name[1] == '.')
				continue;
		}

		safecat(subpath, SUBPATHLEN, node->path, entry.name, 0);

		switch(entry.type) {
			case ENTRY_DIR:
				if((dn = dirnode_create(subpath))) {
					dltmp = NEW(DirListNode);
					dltmp->node = dn;
					dltmp->next = dirlist;
					dirlist = dltmp;
					node->subdir_count++;
				}
				break;
			case ENTRY_FILE:
				if((fn = filenode_create(subpath))) {
					fltmp = NEW(FileListNode);
					fltmp->node = fn;
					fltmp->next = filelist;
					filelist = fltmp;
					node->file_count++;
				}
				break;
		}
	}

	if(node->file_count) {
		node->files = (FileNode *)xmalloc(sizeof(FileNode) * node->file_count);

		for(i = 0; i < node->file_count; i++) {
			node->files[i] = filelist->node;
			fltmp = filelist;
			filelist = fltmp->next;

			xfree(fltmp);
		}
	}

	if(node->subdir_count) {
		node->subdirs = (DirNode *)xmalloc(sizeof(DirNode) * node->subdir_count);

		for(i = 0; i < node->subdir_count; i++) {
			node->subdirs[i] = dirlist->node;
			dltmp = dirlist;
			dirlist = dltmp->next;

			xfree(dltmp);
		}
	}

	sysdir_close(dir);

	return node;

err2:
	sysdir_close(dir);
err1:
#ifndef WIN32
	close(node->fd);
#endif /* WIN32 */
err0:
	xfree(node);
	return 0;
}

static void dirnode_destroy(DirNode node)
{
	unsigned i;

	dirnode_unmonitor(node);

#ifndef NDEBUG
	debug("Removed monitor for: %s", node->path);
#endif
	hash_table_remove(dirnode_table, node->path);

#ifndef WIN32
	close(node->fd);
#endif
	xfree(node->path);
	xfree(node->name);

	for(i = 0; i < node->subdir_count; i++)
		dirnode_destroy(node->subdirs[i]);

	if(node->subdir_count)
		xfree(node->subdirs);

	for(i = 0; i < node->file_count; i++)
		filenode_destroy(node->files[i]);

	if(node->file_count)
		xfree(node->files);

	xfree(node);
}

static void dirnode_event(DirNode node, int recursive)
{
	unsigned i;
	size_t l;
	SysDir dir;
	struct DirEntry entry;
#ifdef WIN32
	char subpath[PDTP_PATH_MAX];
	const unsigned SUBPATHLEN = PDTP_PATH_MAX;
#else
	unsigned dotfiles = 1;
	char subpath[PATH_MAX];
	const unsigned SUBPATHLEN = PATH_MAX;
#endif /* WIN32 */

	HashTable subdir_table, file_table;
	DirListNode dirlist = 0, dltmp;
	FileListNode filelist = 0, fltmp;
	DirNode dn;
	FileNode fn;

#ifndef NDEBUG
	debug("Received directory event for %s", node->path);
#endif 

#ifndef WIN32
	if(config_truth_value("files", "ignore_dotfiles"))
		dotfiles = 0;
#endif /* WIN32 */

	subdir_table = hash_table_create();

	for(i = 0; i < node->subdir_count; i++) 
		hash_table_insert(subdir_table, node->subdirs[i]->name, node->subdirs[i]);

	file_table = hash_table_create();

	for(i = 0; i < node->file_count; i++)
		hash_table_insert(file_table, node->files[i]->name, node->files[i]);

	if(node->subdir_count) 
		xfree(node->subdirs);

	if(node->file_count) 
		xfree(node->files);

	node->subdir_count = 0;
	node->file_count = 0;

	if(!(dir = sysdir_open(node->path)))
		goto done;

	while(!sysdir_get_entry(dir, &entry)) {
		if(*entry.name == '.') {
#ifndef WIN32
			if(!dotfiles)
				continue;
#endif

			l = strlen(entry.name);

			/* Ignore '.' and '..' */
			if(l == 1)
				continue;

			if(l == 2 && entry.name[1] == '.')
				continue;
		}

		safecat(subpath, SUBPATHLEN, node->path, entry.name, 0);

		switch(entry.type) {
			case ENTRY_DIR:
				if(!(dn = (DirNode)hash_table_remove(subdir_table, entry.name)) && !(dn = dirnode_create(subpath)))
					continue;

				dltmp = NEW(DirListNode);
				dltmp->node = dn;
				dltmp->next = dirlist;
				dirlist = dltmp;
				node->subdir_count++;
				break;
			case ENTRY_FILE:
				if(!(fn = (FileNode)hash_table_remove(file_table, entry.name)) && !(fn = filenode_create(subpath)))
					continue;

				fltmp = NEW(FileListNode);
				fltmp->node = fn;
				fltmp->next = filelist;
				filelist = fltmp;
				node->file_count++;
				break;
		}
	}

	sysdir_close(dir);

	if(node->file_count) {
		node->files = (FileNode *)xmalloc(sizeof(FileNode) * node->file_count);

		for(i = 0; i < node->file_count; i++) {
			node->files[i] = filelist->node;
			fltmp = filelist;
			filelist = fltmp->next;

			xfree(fltmp);
		}
	}

	if(node->subdir_count) {
		node->subdirs = (DirNode *)xmalloc(sizeof(DirNode) * node->subdir_count);

		for(i = 0; i < node->subdir_count; i++) {
			/* Execute recursive events */
			if(recursive)
				dirnode_event(dirlist->node, 1);

			node->subdirs[i] = dirlist->node;
			dltmp = dirlist;
			dirlist = dltmp->next;

			xfree(dltmp);
		}
	}

done:
	/* Destroy any unmatched entries */
	hash_table_destroy(subdir_table, (void (*)(void *))dirnode_destroy);
	hash_table_destroy(file_table, (void (*)(void *))filenode_destroy);

	return;
}

static void *vfsmonitor_thread(void *arg)
{
	char path[PDTP_PATH_MAX];

	safecpy(path, (char *)arg, PDTP_PATH_MAX);
	xfree(arg);

	vfsmonitor_event_loop(path);

	return 0;
}

/* XXX Add rehashing support */
static int vfsmonitor_thread_init(const char *path)
{
	Thread thread;
	char *pathptr = xstrdup(path);

	if(!monitor_lock) {
		monitor_lock = lock_create(LOCK_SHARED);
		dirnode_table = hash_table_create();
	}

	if(vfsmonitor_local_init() < 0)
		return -1;

	if(!(thread = thr_create(vfsmonitor_thread, pathptr)))
		return -1;

	return 0;
}

void vfsmonitor_init(void)
{
	base_path_lock = lock_create(LOCK_EXCL);

	if(!(base_path = config_value("files", "base_path"))) {
		log_printf("!!! Warning: files::base_path unset, using '.'");
		base_path = xstrdup(".");
	}

	add_trailing_slash(&base_path);

	if(vfsmonitor_thread_init(base_path) < 0)
		fatal("Unable to configure filesystem monitor for: %s", base_path);
}

void vfsmonitor_destroy()
{
	if(monitor_root)
		dirnode_destroy(monitor_root);

	vfsmonitor_local_shutdown();

	hash_table_destroy(dirnode_table, 0);
	lock_destroy(monitor_lock);

	free(base_path);
	lock_destroy(base_path_lock);
}

static void dirnode_stats(DirNode node, unsigned *files, unsigned *dirs)
{
	unsigned i;

	*files += node->file_count;
	*dirs += node->subdir_count;

	for(i = 0; i < node->subdir_count; i++)
		dirnode_stats(node->subdirs[i], files, dirs);
}

static int vfsmonitor_stats(unsigned *files, unsigned *dirs)
{
	*files = 0;
	*dirs = 1; /* Start at 1 for the parent directory */

	lock_reader_acquire(monitor_lock);

	if(!monitor_root) {
		lock_reader_release(monitor_lock);
		return -1;
	}

	dirnode_stats(monitor_root, files, dirs);
	lock_reader_release(monitor_lock);

	return 0;
}

static void vfs_scan_complete(const char *path)
{
	unsigned files, dirs;

	/* Report statistics after filesystem scan */
	if(vfsmonitor_stats(&files, &dirs) || !files) {
		log_printf("!!! Warning: No files available under %s", path);
		return;
	}

	log_printf("*** Processed '%s': Found %d files in %d directories", path, files, dirs);
}

/* XXX These should really be factored elsewhere... */
char *vfsmonitor_hton_path(const char *path)
{
#ifdef WIN32
	unsigned i;
#endif
	unsigned l;
	char *ret = 0;
	
#ifndef NDEBUG
	debug("Translating '%s'", path);
#endif

	lock_reader_acquire(base_path_lock);
	
	assert(base_path);
	assert((l = strlen(base_path)));

#ifndef NDEBUG
	debug("Comparing '%s' to '%s'", base_path, path);
#endif

	/* Validate that path is actually under the monitor root */
	if(strncmp(base_path, path, l))
		goto done;
	
#ifdef WIN32
	if(base_path[l - 1] == '\\')
		l--;

	ret = xstrdup(path + l);
	
	l = strlen(ret);
	for(i = 0; i < l; i++) {
		if(ret[i] == '\\')
			ret[i] = '/';
	}
#else
	if(base_path[l - 1] == '/')
		l--;

	ret = xstrdup(path + l);
#endif /* WIN32 */

done:
	lock_reader_release(base_path_lock);
	return ret;
}

char *vfsmonitor_ntoh_path(const char *path)
{
#ifdef WIN32
	unsigned i;
	size_t l;
#endif
	char fullpath[PDTP_PATH_MAX];

	/* Check for required leading slash */
	if(*path != '/')
		return 0;

	path++;

	lock_reader_acquire(base_path_lock);
	
	assert(base_path);
	safecat(fullpath, PDTP_PATH_MAX, base_path, path, 0);
	
	lock_reader_release(monitor_lock);

	/* Reverse slashes on Win32 */
#ifdef WIN32
	l = strlen(fullpath);
	for(i = 0; i < l; i++) {
		if(fullpath[i] == '/')
			fullpath[i] = '\\';
	}
#endif

	return xstrdup(fullpath);
}

DirList vfsmonitor_read_dirlist(const char *path)
{
	unsigned i;
	DirNode node;
	DirList dirlist;
	struct DirEntry entry;

	lock_reader_acquire(monitor_lock);

#ifndef NDEBUG
	debug("Listing: %s", path);
#endif

	if(!(node = hash_table_lookup(dirnode_table, path))) {
		lock_reader_release(monitor_lock);
		return 0;
	}

	dirlist = dirlist_create();

	for(i = 0; i < node->subdir_count; i++) {
		entry.type = ENTRY_DIR;
		entry.ts_year = node->subdirs[i]->ts_year;
		entry.ts_ysec = node->subdirs[i]->ts_ysec;
		safecpy(entry.name, node->subdirs[i]->name, PDTP_PATH_MAX);
		entry.length = 0;

		dirlist_add_entry(dirlist, &entry);
	}

	for(i = 0; i < node->file_count; i++) {
		entry.type = ENTRY_FILE;
		entry.ts_year = node->files[i]->ts_year;
		entry.ts_ysec = node->files[i]->ts_ysec;
		safecpy(entry.name, node->files[i]->name, PDTP_PATH_MAX);
		entry.length = node->files[i]->length;

		dirlist_add_entry(dirlist, &entry);
	}

	lock_reader_release(monitor_lock);

	return dirlist;
}
