/**
  @file netevent.c 
  @brief Network event handlers (invoked by dispatcher)
*/
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

/* Common headers */
#include <Config.h>
#include <Dispatcher.h>
#include <auth.h>
#include <debug.h>
#include <log.h>
#include <msgval.h>
#include <socketops.h>
#include <util.h>
#include <xmalloc.h>

/* Local headers */
#include "ServerManager.h"
#include "netevent.h"
#include "servertxn.h"

/* Protocol magic numbers */
#define PDSP_MAGIC      0x50445350
#define PTPP_MAGIC	0x50745050

/** Handle authorization.  Zero on success, -1 on failure */
static int source_auth_handler(int sock)
{
	int ret;
	char *key;
	
	/* Override to allow source connections without authorization */
	if(config_truth_value("servers", "disable_auth") == 1)
		return 1;
	
        /* XXX Only method currently supported.  This interface will change
           in the future to support more methods, such as reading the key
           from a file */
	if(!(key = config_value("servers", "hmac_sha1_passphrase")))
		return 0;

	/* XXX A security audit of this code is in order */
	ret = auth_passphrase(sock, key);
	free(key);

	return ret;
}

static int source_connection_handler(int sock, uint32_t addr)
{
	char *t;
	uint16_t value;
	
	if(source_auth_handler(sock) < 0)
		return -1;
	
	writesock_int32(sock, PDSP_MAGIC);

	/* Minimum protocol version supported, currently hardcoded to 1 */
	writesock_int16(sock, 1);

	/* Maximum protocol version supported, currently hardcoded to 1 */
	writesock_int16(sock, 1);

	/* Read the server's selected protocol version */
	if(readsock_int16(sock, &value) < 0)
		return -1;

	/* Only valid selection at this point */
	if(value != 1)
		return -1;

	/* Add the connection to the server manager */
	servermgr_add_ipv4_connection(sock, addr);

	/* Add the server connection to the dispatcher */
	dispatcher_add(sock, source_transaction_handler);

	log_printf("*** Successfully authenticated server connection from: %s", t = xinet_ntoa(addr));
	xfree(t);

	return 0;
}

static int piecesvr_connection_handler(int sock, uint32_t addr)
{
	return -1;
}

/** Handler for incoming connections */
void incoming_connection_handler(msgval *arg)
{
	Sock sock;
	uint32_t addr, value;
	char *t;

	sock = listener_ipv4_tcp_accept(arg->num, &addr);
	dispatcher_add(arg->num, incoming_connection_handler);

	if(sock < 0)
		return;

	log_printf("+++ Connection from %s", t = xinet_ntoa(addr)); 
	xfree(t);

	/* Read the magic number from the client */
	if(readsock_int32(sock, &value) < 0)
		goto err;

	switch(value) {
		case PDSP_MAGIC:
			if(source_connection_handler(sock, addr) < 0)
				goto err;

			break;
		case PTPP_MAGIC:
			if(piecesvr_connection_handler(sock, addr) < 0)
				goto err;
			
			break;
		default:
			goto err;
	}
	
	return;
err: 
	log_printf("--- %s disconnected", t = xinet_ntoa(addr)); 
	xfree(t);
					
	socket_close(sock);
}

void source_transaction_handler(msgval *arg)
{
        TransactionIn txn;
        int sock = arg->num;

        if(!(txn = transaction_read(sock)) || servertxn_exec(sock, txn) < 0) {
		servermgr_remove_connection(sock);
                return;
        }

        dispatcher_add(sock, source_transaction_handler);
}
