/**
  @file servertxn.c 
  @brief Squall Hub's handlers for server to hub transactions
*/
/*
   Copyright (C)2004 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
*/

#include <machdep.h>

/* Common headers */
#include <Collection.h>
#include <Lock.h>
#include <Transaction.h>
#include <debug.h>
#include <transaction_replies.h>
#include <txid.h>
#include <util.h>
#include <xmalloc.h>

/* Local headers */
#include "DirList.h"
#include "FileManager.h"
#include "VFSMonitor.h"
#include "servertxn.h"

static int txn_fileinfo(int sock, TransactionIn txn);
static int txn_dirlist(int sock, TransactionIn txn);
	
struct th_list_element {
        uint16_t txn_id;
        int (*handler)(int sock, TransactionIn);
};

static struct th_list_element transaction_list[] = {
        { TXN_FILEINFO, txn_fileinfo },
        { TXN_DIRLIST,  txn_dirlist },
        { 0, 0 }
};

static Collection transaction_table = 0;

void servertxn_init()
{
        int i;

        transaction_table = collection_create();

        for(i = 0; transaction_list[i].txn_id; i++)
                collection_insert(transaction_table, (CollectionKey)transaction_list[i].txn_id, (void *)transaction_list[i].handler);
}

void servertxn_destroy()
{
        collection_destroy(transaction_table, 0);
}

int servertxn_exec(int sock, TransactionIn txn)
{ 
        int ret;
        int (*handler)(int, TransactionIn);

        if(!(handler = (int (*)(int, TransactionIn))collection_lookup(transaction_table, (CollectionKey)transaction_id(txn)))) {
                reply_error(sock, transaction_serial(txn), "Invalid transaction type (0x%x)", transaction_id(txn));
                return 0;
        }

        ret = handler(sock, txn);
        transaction_in_destroy(txn);

        return ret;
}

static int txn_fileinfo(int sock, TransactionIn txn)
{
	int i;
	char *pdtp_path, *host_path;
	FileEntry entry;
	FileHashes hashes;
	TransactionOut reply;

        if((i = transaction_object_index(txn, OBJ_PATH)) < 0) {
                reply_error(sock, transaction_serial(txn), "No path given");
                return 0;
        }
 
        pdtp_path = transaction_object_string(txn, i);
        host_path = vfsmonitor_ntoh_path(pdtp_path);
        xfree(pdtp_path);
 
        if(!host_path) {
                reply_error(sock, transaction_serial(txn), "Invalid path");
                return 0;
        }

#ifdef DEBUG
	debug("Request for file info for: %s", host_path);
#endif

	entry = filemgr_get_entry_by_path(host_path, LOCK_SHARED);
	xfree(host_path);

	if(!entry) { 
		reply_error(sock, transaction_serial(txn), "File not found");
		return 0;
	}

	if(!(hashes = fileentry_get_hashes(entry))) {
		filemgr_return_entry(entry);
		reply_error(sock, transaction_serial(txn), "Error: This file is not yet hashed.  It will be available shortly.");

		return 0;
	}

	reply = transaction_create(TXN_REPLY, TXN_FILEINFO, 4);
	transaction_set_serial(reply, transaction_serial(txn));

	transaction_add_int64(reply, OBJ_FILELEN, fileentry_get_length(entry));
	transaction_add_int32(reply, OBJ_PIECELEN, filehashes_get_piece_length(hashes));
	filehashes_add_to_transaction(reply, hashes);

	filehashes_destroy(hashes);
	filemgr_return_entry(entry);

        transaction_write(sock, reply);
        transaction_out_destroy(reply);
	
	return 0;
}

static int txn_dirlist(int sock, TransactionIn txn)
{
	int i;
	char *pdtp_path, *host_path;
	DirList dirlist;
	TransactionOut reply;
	struct DirEntry entry;

	if((i = transaction_object_index(txn, OBJ_PATH)) < 0) {
		reply_error(sock, transaction_serial(txn), "No path given");
		return 0;
	}

	pdtp_path = transaction_object_string(txn, i);
	host_path = vfsmonitor_ntoh_path(pdtp_path);
	xfree(pdtp_path);

	if(!host_path) {
		reply_error(sock, transaction_serial(txn), "Invalid path");
		return 0;
	}

	dirlist = vfsmonitor_read_dirlist(host_path);
	xfree(host_path);

	if(!dirlist) {
		reply_error(sock, transaction_serial(txn), "Invalid path");
		return 0;
	}

	reply = transaction_create(TXN_REPLY, TXN_DIRLIST, dirlist_count(dirlist));
	transaction_set_serial(reply, transaction_serial(txn));

	while(!dirlist_get_entry(dirlist, &entry)) {
		switch(entry.type) {
			case ENTRY_FILE:
				transaction_object_pack(reply, OBJ_FILE, "248s", entry.ts_year, entry.ts_ysec, entry.length, entry.name);
				break;
			case ENTRY_DIR:
				transaction_object_pack(reply, OBJ_SUBDIR, "24s", entry.ts_year, entry.ts_ysec, entry.name); 
				break;
		}
	}

	dirlist_destroy(dirlist);

	transaction_write(sock, reply);
	transaction_out_destroy(reply);
		
	return 0;
}
