/**
  @file main.c 
  @brief Squall PDTP piece proxy server main program
 */
/*
   Copyright (C)2003-04 Anthony Arcieri 
   All rights reserved.

   For licensing information, please see the file LICENSE 
 */

#include <machdep.h>

#ifdef HAVE_PTHREADS
#include <signal.h>
#include <pthread.h>
#endif

/* Common headers */
#include <ComponentManager.h>
#include <Config.h>
#include <Dispatcher.h>
#include <PoolCleaner.h>
#include <WorkerManager.h>
#include <log.h>
#include <util.h>

/* Local headers */
#include "config_keys.h"
#include "squallproxy.h"

static void init_logging(void);
static void init_config(void);
#ifdef WIN32
static void init_sockets(void);
#else
static void init_signals(void);
#endif /* WIN32 */

/** Paths to search for configuration file */
#ifdef WIN32
static const char *config_file_path[] = {
        "squallproxy.ini",
        "c:\\windows\\squallproxy.ini",
        "c:\\winnt\\squallproxy.ini",
        0
};

#else

static const char *config_file_path[] = { 
        "/var/squall/squallproxy.conf",
        "/etc/squallproxy.conf", 
        "/usr/local/etc/squallproxy.conf",
        "squallproxy.conf",
        0
};
#endif

static void component_init(void)
{
	Component Logging;
	Component PoolCleaner;
	Component Dispatcher;
	Component Configuration;

#ifdef WIN32
        /* Win32-specific components */ 
        Component Sockets;
#else
        /* POSIX-specific components */
        Component Signals;
        Component WorkerManager;
#endif /* WIN32 */

	componentmgr_init();

	Configuration = component_create(init_config, init_config, config_destroy);

#ifdef WIN32
        Sockets = component_create(init_sockets, 0, 0);
#else
        Signals = component_create(init_signals, 0, 0);
#endif

        Logging = component_create(init_logging, 0, log_shutdown);

        PoolCleaner = component_create(poolcleaner_init, 0, poolcleaner_destroy);

#ifndef WIN32
        WorkerManager = component_create(workermgr_init, 0, workermgr_destroy);
        component_register_dependency(WorkerManager, Configuration);
#endif /* WIN32 */

        Dispatcher = component_create(dispatcher_init, 0, dispatcher_destroy);
        /* component_register_dependency(Dispatcher, TransactionHandler); */
#ifndef WIN32
        component_register_dependency(Dispatcher, WorkerManager);
#endif /* WIN32 */

	componentmgr_init_components();
}

#ifdef WIN32
static void init_sockets()
{
        WSADATA data;
 
        /* Initialize Winsock */
        if(WSAStartup(MAKEWORD(2, 0), &data))
                fatal("!!! Couldn't initialize Winsock");
}
#else
static void init_signals(void)
{
        sigset_t sigset;

        /* Block signals we plan to handle or want to ignore */
        sigemptyset(&sigset);
        sigaddset(&sigset, SIGHUP);
        sigaddset(&sigset, SIGINT);
        sigaddset(&sigset, SIGTERM);
        sigaddset(&sigset, SIGPIPE);
        
        if(pthread_sigmask(SIG_BLOCK, &sigset, 0) < 0) {
                log_error("pthread_sigmask");
                fatal("!!! Couldn't set signal mask");
        }
}
#endif /* WIN32 */

static void init_logging(void)
{
        struct systime st;

        log_init();

        get_systime(&st);
        log_printf("*** %s %s started on %04d-%02d-%02d", PACKAGE, VERSION, st.year, st.month, st.day);
#ifndef WIN32
        log_printf("*** Built for %s on %s", PLATFORM, BUILDDATE);
#endif /* WIN32 */
}

/** Small wrapper for config_init, handles paths */
static void init_config(void)
{
        int i;

        for(i = 0; config_file_path[i]; i++) { 
                if(!config_init(config_file_path[i], config_keys)) {
                        log_printf("*** Configuration successfully loaded from '%s'", config_file_path[i]);
                        return;
                }
        }

        log_printf("!!! Warning: Couldn't open configuration file, using default settings"); 
}

int main(int argc, char **argv)
{
	/* Initialize all system components */
	component_init();

	return 0;
}
